# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import itertools

import numpy as np
import pytest

import sisl as si
import sisl.geom as sisl_geom
from sisl import (
    Atom,
    Cube,
    Geometry,
    Lattice,
    SislDeprecation,
    SislError,
    SislWarning,
    Sphere,
)

pytestmark = [pytest.mark.geom, pytest.mark.geometry]


@pytest.fixture
def setup():
    class t:
        def __init__(self):
            bond = 1.42
            sq3h = 3.0**0.5 * 0.5
            self.lattice = Lattice(
                np.array(
                    [[1.5, sq3h, 0.0], [1.5, -sq3h, 0.0], [0.0, 0.0, 10.0]], np.float64
                )
                * bond,
                nsc=[3, 3, 1],
            )
            C = Atom(Z=6, R=[bond * 1.01] * 2)
            self.g = Geometry(
                np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0]], np.float64) * bond,
                atoms=C,
                lattice=self.lattice,
            )

            self.mol = Geometry([[i, 0, 0] for i in range(10)], lattice=[50])

    return t()


class TestGeometry:
    def test_objects(self, setup):
        str(setup.g)
        assert len(setup.g) == 2
        assert len(setup.g.xyz) == 2
        assert np.allclose(setup.g[0], np.zeros([3]))
        assert np.allclose(setup.g[None, 0], setup.g.xyz[:, 0])

        i = 0
        for ia in setup.g:
            i += 1
        assert i == len(setup.g)
        assert setup.g.no_s == 2 * len(setup.g) * np.prod(setup.g.lattice.nsc)

    def test_properties(self, setup):
        assert 2 == len(setup.g)
        assert 2 == setup.g.na
        assert 3 * 3 == setup.g.n_s
        assert 2 * 3 * 3 == setup.g.na_s
        assert 2 * 2 == setup.g.no
        assert 2 * 2 * 3 * 3 == setup.g.no_s

    def test_iter1(self, setup):
        i = 0
        for ia in setup.g:
            i += 1
        assert i == 2

    def test_iter2(self, setup):
        for ia in setup.g:
            assert np.allclose(setup.g[ia], setup.g.xyz[ia, :])

    def test_iter3(self, setup):
        i = 0
        for ia, io in setup.g.iter_orbitals(0):
            assert ia == 0
            assert io < 2
            i += 1
        for ia, io in setup.g.iter_orbitals(1):
            assert ia == 1
            assert io < 2
            i += 1
        assert i == 4
        i = 0
        for ia, io in setup.g.iter_orbitals():
            assert ia in [0, 1]
            assert io < 2
            i += 1
        assert i == 4

        i = 0
        for ia, io in setup.g.iter_orbitals(1, local=False):
            assert ia == 1
            assert io >= 2
            i += 1
        assert i == 2

    def test_tile0(self, setup):
        with pytest.raises(ValueError):
            t = setup.g.tile(0, 0)

    def test_tile1(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[0, :] *= 2
        t = setup.g.tile(2, 0)
        assert np.allclose(cell, t.lattice.cell)
        cell[1, :] *= 2
        t = t.tile(2, 1)
        assert np.allclose(cell, t.lattice.cell)
        cell[2, :] *= 2
        t = t.tile(2, 2)
        assert np.allclose(cell, t.lattice.cell)

    def test_tile2(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[:, :] *= 2
        t = setup.g.tile(2, 0).tile(2, 1).tile(2, 2)
        assert np.allclose(cell, t.lattice.cell)

    def test_sort(self, setup):
        t = setup.g.tile(2, 0).tile(2, 1).tile(2, 2)
        ts = t.sort()
        t = setup.g.tile(2, 1).tile(2, 2).tile(2, 0)
        tS = t.sort()
        assert np.allclose(ts.xyz, tS.xyz)

    def test_tile3(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[:, :] *= 2
        t1 = setup.g * 2
        cell = np.copy(setup.g.lattice.cell)
        cell[0, :] *= 2
        t1 = setup.g * (2, 0)
        assert np.allclose(cell, t1.lattice.cell)
        t = setup.g * ((2, 0), "tile")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)
        cell[1, :] *= 2
        t1 = t * (2, 1)
        assert np.allclose(cell, t1.lattice.cell)
        t = t * ((2, 1), "tile")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)
        cell[2, :] *= 2
        t1 = t * (2, 2)
        assert np.allclose(cell, t1.lattice.cell)
        t = t * ((2, 2), "tile")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)

        # Full
        t = setup.g * [2, 2, 2]
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)
        t = setup.g * ([2, 2, 2], "t")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)

    def test_tile4(self, setup):
        t1 = setup.g.tile(2, 0).tile(2, 2)
        t = setup.g * ([2, 0], "t") * [2, 2]
        assert np.allclose(t1.xyz, t.xyz)

    def test_tile5(self, setup):
        t = setup.g.tile(2, 0).tile(2, 2)
        assert np.allclose(t[: len(setup.g), :], setup.g.xyz)

    def test_repeat0(self, setup):
        with pytest.raises(ValueError):
            t = setup.g.repeat(0, 0)

    def test_repeat1(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[0, :] *= 2
        t = setup.g.repeat(2, 0)
        assert np.allclose(cell, t.lattice.cell)
        cell[1, :] *= 2
        t = t.repeat(2, 1)
        assert np.allclose(cell, t.lattice.cell)
        cell[2, :] *= 2
        t = t.repeat(2, 2)
        assert np.allclose(cell, t.lattice.cell)

    def test_repeat2(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[:, :] *= 2
        t = setup.g.repeat(2, 0).repeat(2, 1).repeat(2, 2)
        assert np.allclose(cell, t.lattice.cell)

    def test_repeat3(self, setup):
        cell = np.copy(setup.g.lattice.cell)
        cell[0, :] *= 2
        t1 = setup.g.repeat(2, 0)
        assert np.allclose(cell, t1.lattice.cell)
        t = setup.g * ((2, 0), "repeat")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)
        cell[1, :] *= 2
        t1 = t.repeat(2, 1)
        assert np.allclose(cell, t1.lattice.cell)
        t = t * ((2, 1), "r")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)
        cell[2, :] *= 2
        t1 = t.repeat(2, 2)
        assert np.allclose(cell, t1.lattice.cell)
        t = t * ((2, 2), "repeat")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)

        # Full
        t = setup.g * ([2, 2, 2], "r")
        assert np.allclose(cell, t.lattice.cell)
        assert np.allclose(t1.xyz, t.xyz)

    def test_repeat4(self, setup):
        t1 = setup.g.repeat(2, 0).repeat(2, 2)
        t = setup.g * ([2, 0], "repeat") * ([2, 2], "r")
        assert np.allclose(t1.xyz, t.xyz)

    def test_repeat5(self, setup):
        t = setup.g.repeat(2, 0).repeat(2, 2)
        assert np.allclose(t.xyz[::4, :], setup.g.xyz)

    def test_a2o1(self, setup):
        assert 0 == setup.g.a2o(0)
        assert setup.g.atoms[0].no == setup.g.a2o(1)
        assert setup.g.no == setup.g.a2o(setup.g.na)

    def test_sub1(self, setup):
        assert len(setup.g.sub([0])) == 1
        assert len(setup.g.sub([0, 1])) == 2
        assert len(setup.g.sub([-1])) == 1

        assert np.allclose(
            setup.g.sub([0]).xyz, setup.g.sub(np.array([True, False])).xyz
        )

    def test_sub2(self, setup):
        assert len(setup.g.sub(range(1))) == 1
        assert len(setup.g.sub(range(2))) == 2

    def test_fxyz(self, setup):
        fxyz = setup.g.fxyz
        assert np.allclose(fxyz, [[0, 0, 0], [1.0 / 3, 1.0 / 3, 0]])
        assert np.allclose(np.dot(fxyz, setup.g.cell), setup.g.xyz)

    def test_axyz(self, setup):
        g = setup.g
        assert np.allclose(g[:], g.xyz[:])
        assert np.allclose(g[0], g.xyz[0, :])
        assert np.allclose(g[2], g.axyz(2))
        isc = g.a2isc(2)
        off = g.lattice.offset(isc)
        assert np.allclose(g.xyz[0] + off, g.axyz(2))
        assert np.allclose(g.xyz[0] + off, g.axyz(0, isc))
        assert np.allclose(g.xyz[0] + off, g.axyz(isc=isc)[0])

        # Also check multidimensional things
        assert g.axyz([[0], [1]]).shape == (2, 1, 3)

    def test_atranspose_indices(self, setup):
        g = setup.g
        # All supercell indices
        ia2 = np.arange(g.na * g.n_s)
        ja2, ja1 = g.a2transpose(0, ia2)
        assert (ja2 < g.na).sum() == ja2.size
        assert (ja1 % g.na == 0).sum() == ja1.size

        IA1, IA2 = g.a2transpose(ja2, ja1)
        assert np.all(IA1 == 0)
        assert np.all(IA2 == ia2)

    def test_otranspose_indices(self, setup):
        g = setup.g
        # All supercell indices
        io2 = np.arange(g.no * g.n_s)
        jo2, jo1 = g.o2transpose(0, io2)
        assert (jo2 < g.no).sum() == jo2.size
        assert (jo1 % g.no == 0).sum() == jo1.size

        IO1, IO2 = g.o2transpose(jo2, jo1)
        assert np.all(IO1 == 0)
        assert np.all(IO2 == io2)

    def test_auc2sc(self, setup):
        g = setup.g
        # All supercell indices
        asc = g.auc2sc(0)
        assert asc.size == g.n_s
        assert (asc % g.na == 0).sum() == g.n_s

    def test_ouc2sc(self, setup):
        g = setup.g
        # All supercell indices
        asc = g.ouc2sc(0)
        assert asc.size == g.n_s
        assert (asc % g.no == 0).sum() == g.n_s

    def test_rij1(self, setup):
        assert np.allclose(setup.g.rij(0, 1), 1.42)
        assert np.allclose(setup.g.rij(0, [0, 1]), [0.0, 1.42])

    def test_orij1(self, setup):
        assert np.allclose(setup.g.orij(0, 2), 1.42)
        assert np.allclose(setup.g.orij(0, [0, 2]), [0.0, 1.42])

    def test_Rij1(self, setup):
        assert np.allclose(setup.g.Rij(0, 1), [1.42, 0, 0])

    def test_oRij1(self, setup):
        assert np.allclose(setup.g.oRij(0, 1), [0.0, 0, 0])
        assert np.allclose(setup.g.oRij(0, 2), [1.42, 0, 0])
        assert np.allclose(
            setup.g.oRij(0, [0, 1, 2]), [[0.0, 0, 0], [0.0, 0, 0], [1.42, 0, 0]]
        )
        assert np.allclose(setup.g.oRij(0, 2), [1.42, 0, 0])

    def test_untile_warns(self, setup):
        with pytest.warns(SislWarning) as warns:
            assert len(setup.g.untile(1, 1)) == 2
            assert len(setup.g.untile(2, 1)) == 1
            assert len(setup.g.untile(2, 1, 1)) == 1
        assert len(warns) == 2

    def test_untile_check_same(self, setup):
        with pytest.warns(SislWarning) as warns:
            c1 = setup.g.untile(2, 1)
            c2 = setup.g.untile(2, 1, 1)
        assert len(warns) == 2
        assert np.allclose(c1.xyz[0, :], setup.g.xyz[0, :])
        assert np.allclose(c2.xyz[0, :], setup.g.xyz[1, :])

    def test_untile_algo(self, setup):
        nr = range(2, 5)
        g = setup.g.copy()
        for x in nr:
            gx = g.tile(x, 0)
            for y in nr:
                gy = gx.tile(y, 1)
                for z in nr:
                    gz = gy.tile(z, 2)
                    G = gz.untile(z, 2)
                    assert np.allclose(G.xyz, gy.xyz)
                    assert np.allclose(G.cell, gy.cell)
                G = gy.untile(y, 1)
                assert np.allclose(G.xyz, gx.xyz)
                assert np.allclose(G.cell, gx.cell)
            G = gx.untile(x, 0)
            assert np.allclose(G.xyz, g.xyz)
            assert np.allclose(G.cell, g.cell)

    def test_unrepeat_algo(self, setup):
        nr = range(2, 5)
        g = setup.g.copy()
        for x in nr:
            gx = g.repeat(x, 0)
            for y in nr:
                gy = gx.repeat(y, 1)
                for z in nr:
                    gz = gy.repeat(z, 2)
                    G = gz.unrepeat(z, 2)
                    assert np.allclose(G.xyz, gy.xyz)
                    assert np.allclose(G.cell, gy.cell)
                G = gy.unrepeat(y, 1)
                assert np.allclose(G.xyz, gx.xyz)
                assert np.allclose(G.cell, gx.cell)
            G = gx.unrepeat(x, 0)
            assert np.allclose(G.xyz, g.xyz)
            assert np.allclose(G.cell, g.cell)

    def test_remove1(self, setup):
        assert len(setup.g.remove([0])) == 1
        assert len(setup.g.remove([])) == 2
        assert len(setup.g.remove([-1])) == 1
        assert len(setup.g.remove([-0])) == 1

    def test_remove2(self, setup):
        assert len(setup.g.remove(range(1))) == 1
        assert len(setup.g.remove(range(0))) == 2

    def test_copy(self, setup):
        assert setup.g == setup.g.copy()

    def test_nsc1(self, setup):
        lattice = setup.g.lattice.copy()
        nsc = np.copy(lattice.nsc)
        lattice.set_nsc([5, 5, 0])
        assert np.allclose([5, 5, 1], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)

    def test_nsc2(self, setup):
        lattice = setup.g.lattice.copy()
        nsc = np.copy(lattice.nsc)
        lattice.set_nsc([0, 1, 0])
        assert np.allclose([1, 1, 1], lattice.nsc)
        assert len(lattice.sc_off) == np.prod(lattice.nsc)

    def test_rotation1(self, setup):
        rot = setup.g.rotate(180, [0, 0, 1], what="xyz+abc")
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = setup.g.rotate(np.pi, [0, 0, 1], rad=True, what="xyz+abc")
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = rot.rotate(180, "z", what="xyz+abc")
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(rot.xyz, setup.g.xyz)

    def test_rotation2(self, setup):
        rot = setup.g.rotate(180, "z", what="abc")
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(rot.xyz, setup.g.xyz)

        rot = setup.g.rotate(np.pi, [0, 0, 1], rad=True, what="abc")
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(-rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(rot.xyz, setup.g.xyz)

        rot = rot.rotate(180, [0, 0, 1], what="abc")
        rot.lattice.cell[2, 2] *= -1
        assert np.allclose(rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(rot.xyz, setup.g.xyz)

    def test_rotation3(self, setup):
        rot = setup.g.rotate(180, [0, 0, 1], what="xyz")
        assert np.allclose(rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = setup.g.rotate(np.pi, [0, 0, 1], rad=True, what="xyz")
        assert np.allclose(rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(-rot.xyz, setup.g.xyz)

        rot = rot.rotate(180, "z", what="xyz")
        assert np.allclose(rot.lattice.cell, setup.g.lattice.cell)
        assert np.allclose(rot.xyz, setup.g.xyz)

    def test_rotation4(self, setup):
        ref = setup.g.tile(2, 0).tile(2, 1)

        rot = ref.rotate(10, "z", atoms=1)
        assert not np.allclose(ref.xyz[1], rot.xyz[1])

        rot = ref.rotate(10, "z", atoms=[1, 2])
        assert not np.allclose(ref.xyz[1], rot.xyz[1])
        assert not np.allclose(ref.xyz[2], rot.xyz[2])
        assert np.allclose(ref.xyz[3], rot.xyz[3])

        rot = ref.rotate(10, "z", atoms=[1, 2], what="y")
        assert ref.xyz[1, 0] == rot.xyz[1, 0]
        assert ref.xyz[1, 1] != rot.xyz[1, 1]
        assert ref.xyz[1, 2] == rot.xyz[1, 2]

        rot = ref.rotate(10, "z", atoms=[1, 2], what="xy", origin=ref.xyz[2])
        assert ref.xyz[1, 0] != rot.xyz[1, 0]
        assert ref.xyz[1, 1] != rot.xyz[1, 1]
        assert ref.xyz[1, 2] == rot.xyz[1, 2]
        assert np.allclose(ref.xyz[2], rot.xyz[2])

    def test_translate(self, setup):
        t = setup.g.translate([0, 0, 1])
        assert np.allclose(setup.g.xyz[:, 0], t.xyz[:, 0])
        assert np.allclose(setup.g.xyz[:, 1], t.xyz[:, 1])
        assert np.allclose(setup.g.xyz[:, 2] + 1, t.xyz[:, 2])
        t = setup.g.move([0, 0, 1])
        assert np.allclose(setup.g.xyz[:, 0], t.xyz[:, 0])
        assert np.allclose(setup.g.xyz[:, 1], t.xyz[:, 1])
        assert np.allclose(setup.g.xyz[:, 2] + 1, t.xyz[:, 2])

    def test_iter_block1(self, setup):
        for i, iaaspec in enumerate(setup.g.iter_species()):
            ia, a, spec = iaaspec
            assert i == ia
            assert setup.g.atoms[ia] == a
        for ia, a, spec in setup.g.iter_species([1]):
            assert 1 == ia
            assert setup.g.atoms[ia] == a
        for ia in setup.g:
            assert ia >= 0
        i = 0
        for ias, idx in setup.g.iter_block():
            for ia in ias:
                i += 1
        assert i == len(setup.g)

        i = 0
        for ias, idx in setup.g.iter_block(atoms=1):
            for ia in ias:
                i += 1
        assert i == 1

    @pytest.mark.slow
    def test_iter_block2(self, setup):
        g = setup.g.tile(30, 0).tile(30, 1)
        i = 0
        for ias, _ in g.iter_block():
            i += len(ias)
        assert i == len(g)

    def test_iter_shape1(self, setup):
        i = 0
        for ias, _ in setup.g.iter_block(method="sphere"):
            i += len(ias)
        assert i == len(setup.g)
        i = 0
        for ias, _ in setup.g.iter_block(method="cube"):
            i += len(ias)
        assert i == len(setup.g)

    @pytest.mark.slow
    def test_iter_shape2(self, setup):
        g = setup.g.tile(30, 0).tile(30, 1)
        i = 0
        for ias, _ in g.iter_block(method="sphere"):
            i += len(ias)
        assert i == len(g)
        i = 0
        for ias, _ in g.iter_block(method="cube"):
            i += len(ias)
        assert i == len(g)
        i = 0
        for ias, _ in g.iter_block_shape(Cube(g.maxR() * 20)):
            i += len(ias)
        assert i == len(g)

    @pytest.mark.slow
    def test_iter_shape3(self, setup):
        g = setup.g.tile(50, 0).tile(50, 1)
        i = 0
        for ias, _ in g.iter_block(method="sphere"):
            i += len(ias)
        assert i == len(g)
        i = 0
        for ias, _ in g.iter_block(method="cube"):
            i += len(ias)
        assert i == len(g)
        i = 0
        for ias, _ in g.iter_block_shape(Sphere(g.maxR() * 20)):
            i += len(ias)
        assert i == len(g)

    def test_swap(self, setup):
        s = setup.g.swap(0, 1)
        for i in [0, 1, 2]:
            assert np.allclose(setup.g.xyz[::-1, i], s.xyz[:, i])

    def test_append1(self, setup):
        for axis in [0, 1, 2]:
            s = setup.g.append(setup.g, axis)
            assert len(s) == len(setup.g) * 2
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            s = setup.g.prepend(setup.g, axis)
            assert len(s) == len(setup.g) * 2
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            s = setup.g.append(setup.g.lattice, axis)
            assert len(s) == len(setup.g)
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            s = setup.g.prepend(setup.g.lattice, axis)
            assert len(s) == len(setup.g)
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)
            assert np.allclose(s.cell[axis, :], setup.g.cell[axis, :] * 2)

    def test_append_raise_valueerror(self, setup):
        with pytest.raises(ValueError):
            s = setup.g.append(setup.g, 0, offset="not")

    def test_prepend_raise_valueerror(self, setup):
        with pytest.raises(ValueError):
            s = setup.g.prepend(setup.g, 0, offset="not")

    def test_append_prepend_offset(self, setup):
        for axis in [0, 1, 2]:
            t = setup.g.lattice.cell[axis, :].copy()
            t *= 10.0 / (t**2).sum() ** 0.5
            s1 = setup.g.copy()
            s2 = setup.g.translate(t)

            S = s1.append(s2, axis, offset="min")
            s = setup.g.append(setup.g, axis)

            assert np.allclose(s.cell[axis, :], S.cell[axis, :])
            assert np.allclose(s.xyz, S.xyz)

            P = s2.prepend(s1, axis, offset="min")
            p = setup.g.prepend(setup.g, axis)

            assert np.allclose(p.cell[axis, :], P.cell[axis, :])
            assert np.allclose(p.xyz, P.xyz)

    @pytest.mark.parametrize("a,b", [[0, 1], [0, 2], [1, 2]])
    def test_swapaxes_lattice_vectors(self, setup, a, b):
        s = setup.g.swapaxes(a, b)
        assert np.allclose(setup.g.xyz, s.xyz)

    @pytest.mark.parametrize("a,b", [[0, 1], [0, 2], [1, 2]])
    def test_swapaxes_lattice_xyz(self, setup, a, b):
        g = setup.g
        s = g.swapaxes(a, b, what="xyz")
        idx = [0, 1, 2]
        idx[a], idx[b] = idx[b], idx[a]
        assert np.allclose(g.cell[:, idx], s.cell)
        assert np.allclose(g.xyz[:, idx], s.xyz)

    def test_swapaxes_complicated(self, setup):
        # swap a couple of lattice vectors and cartesian coordinates
        a = "azby"
        b = "bxcz"
        # this will result in
        # 0. abc, xyz
        # 1. bac, xyz
        # 2. bac, zyx
        # 3. bca, zyx
        # 4. bca, zxy
        sab = setup.g.swapaxes(a, b)
        idx_abc = [1, 2, 0]
        idx_xyz = [2, 0, 1]
        assert np.allclose(setup.g.xyz[:, idx_xyz], sab.xyz)
        assert np.allclose(setup.g.cell[idx_abc][:, idx_xyz], sab.cell)

    def test_center(self, setup):
        g = setup.g.copy()
        assert np.allclose(g[1], g.center(atoms=[1]))
        assert np.allclose(np.mean(g.xyz, axis=0), g.center())
        # in this case the pbc COM is equivalent to the simple one
        assert np.allclose(g.center(what="mass"), g.center(what="mass:pbc"))
        assert np.allclose(g.center(what="mm:xyz"), g.center(what="mm(xyz)"))

    def test_center_raise(self, setup):
        with pytest.raises(ValueError):
            al = setup.g.center(what="unknown")

    def test___add1__(self, setup):
        n = len(setup.g)
        double = setup.g + setup.g + setup.g.lattice
        assert len(double) == n * 2
        assert np.allclose(setup.g.cell * 2, double.cell)
        assert np.allclose(setup.g.xyz[:n, :], double.xyz[:n, :])

        double = (setup.g, 1) + setup.g
        d = setup.g.prepend(setup.g, 1)
        assert len(double) == n * 2
        assert np.allclose(setup.g.cell[::2, :], double.cell[::2, :])
        assert np.allclose(double.xyz, d.xyz)

        double = setup.g + (setup.g, 1)
        d = setup.g.append(setup.g, 1)
        assert len(double) == n * 2
        assert np.allclose(setup.g.cell[::2, :], double.cell[::2, :])
        assert np.allclose(double.xyz, d.xyz)

    def test___add2__(self, setup):
        g1 = setup.g.rotate(15, setup.g.cell[2, :])
        g2 = setup.g.rotate(30, setup.g.cell[2, :])

        assert g1 != g2
        assert g1 + g2 == g1.add(g2)
        assert g1 + g2 != g2.add(g1)
        assert g2 + g1 == g2.add(g1)
        assert g2 + g1 != g1.add(g2)
        for i in range(3):
            assert g1 + (g2, i) == g1.append(g2, i)
            assert (g1, i) + g2 == g2.append(g1, i)

    def test___mul__(self, setup):
        g = setup.g.copy()
        assert g * 2 == g.tile(2, 0).tile(2, 1).tile(2, 2)
        assert g * [2, 1] == g.tile(2, 1)
        assert [2, 1] * g == g.repeat(2, 1)
        assert g * (2, 2, 2) == g.tile(2, 0).tile(2, 1).tile(2, 2)
        assert g * [1, 2, 2] == g.tile(1, 0).tile(2, 1).tile(2, 2)
        assert g * [1, 3, 2] == g.tile(1, 0).tile(3, 1).tile(2, 2)
        assert g * ([1, 3, 2], "r") == g.repeat(1, 0).repeat(3, 1).repeat(2, 2)
        assert g * ([1, 3, 2], "repeat") == g.repeat(1, 0).repeat(3, 1).repeat(2, 2)
        assert g * ([1, 3, 2], "tile") == g.tile(1, 0).tile(3, 1).tile(2, 2)
        assert g * ([1, 3, 2], "t") == g.tile(1, 0).tile(3, 1).tile(2, 2)
        assert g * ([3, 2], "t") == g.tile(3, 2)
        assert g * ([3, 2], "r") == g.repeat(3, 2)

    def test_add(self, setup):
        double = setup.g.add(setup.g)
        assert len(double) == len(setup.g) * 2
        assert np.allclose(setup.g.cell, double.cell)
        double = setup.g.add(setup.g).add(setup.g.lattice)
        assert len(double) == len(setup.g) * 2
        assert np.allclose(setup.g.cell * 2, double.cell)

    def test_add_vacuum(self, setup):
        g = setup.g
        added = setup.g.add_vacuum(g.cell[2, 2], 2)
        assert len(added) == len(g)
        assert np.allclose(
            added.lattice.length, g.lattice.length + [0, 0, g.cell[2, 2]]
        )
        assert np.allclose(added.xyz, g.xyz)
        added = g.add_vacuum(g.cell[2, 2], 2, offset=(1, 2, 3))
        assert len(added) == len(g)
        assert np.allclose(
            added.lattice.length, g.lattice.length + [0, 0, g.cell[2, 2]]
        )
        assert np.allclose(added.xyz, g.xyz + [1, 2, 3])

    def test_insert(self, setup):
        double = setup.g.insert(0, setup.g)
        assert len(double) == len(setup.g) * 2
        assert np.allclose(setup.g.cell, double.cell)

    def test_a2o(self, setup):
        # There are 2 orbitals per C atom
        assert setup.g.a2o(1) == setup.g.atoms[0].no
        assert np.all(setup.g.a2o(1, True) == [2, 3])
        setup.g.reorder()

    def test_o2a(self, setup):
        g = setup.g
        # There are 2 orbitals per C atom
        assert g.o2a(2) == 1
        assert g.o2a(3) == 1
        assert g.o2a(4) == 2
        assert np.all(g.o2a([0, 2, 4]) == [0, 1, 2])
        assert np.all(g.o2a([[0], [2], [4]]) == [[0], [1], [2]])
        assert np.all(g.o2a([[0], [2], [4]], unique=True) == [0, 1, 2])

    def test_angle(self, setup):
        # There are 2 orbitals per C atom
        g = Geometry([[0] * 3, [1, 0, 0]])
        cell = g.cell.copy()
        g.angle([0])
        g.angle([0], ref=1)
        g.angle([0], dir=1)
        assert np.allclose(g.cell[1], cell[1])

    def test_dihedral(self):
        g = sisl_geom.graphene() * (2, 2, 1)
        g.xyz[-1, 2] = 1
        assert g.dihedral(range(4)) == 180
        assert g.dihedral(range(4), rad=True) == np.pi
        assert g.dihedral(range(4, 8)) == pytest.approx(-140.88304639377796)
        assert np.allclose(
            g.dihedral([range(4), range(1, 5), range(4, 8)]),
            [180, 0, -140.88304639377796],
        )

    def test_2uc(self, setup):
        # functions for any-thing to UC
        g = setup.g
        assert g.asc2uc(2) == 0
        assert np.all(g.asc2uc([2, 3]) == [0, 1])
        assert g.asc2uc(2) == 0
        assert np.all(g.asc2uc([2, 3]) == [0, 1])
        assert g.osc2uc(4) == 0
        assert g.osc2uc(5) == 1
        assert np.all(g.osc2uc([4, 5]) == [0, 1])

    def test_2uc_many_axes(self, setup):
        # 2 orbitals per atom
        g = setup.g
        # functions for any-thing to SC
        idx = [[1], [2]]
        assert np.all(g.asc2uc(idx) == [[1], [0]])
        idx = [[2], [4]]
        assert np.all(g.osc2uc(idx) == [[2], [0]])

    def test_2sc(self, setup):
        # functions for any-thing to SC
        g = setup.g
        c = g.cell

        # check indices
        assert np.all(g.a2isc([1, 2]) == [[0, 0, 0], [-1, -1, 0]])
        assert np.all(g.a2isc(2) == [-1, -1, 0])
        assert np.allclose(g.a2sc(2), -c[0, :] - c[1, :])
        assert np.all(g.o2isc([1, 5]) == [[0, 0, 0], [-1, -1, 0]])
        assert np.all(g.o2isc(5) == [-1, -1, 0])
        assert np.allclose(g.o2sc(5), -c[0, :] - c[1, :])

        # Check off-sets
        assert np.allclose(setup.g.a2sc([1, 2]), [[0.0, 0.0, 0.0], -c[0, :] - c[1, :]])
        assert np.allclose(setup.g.o2sc([1, 5]), [[0.0, 0.0, 0.0], -c[0, :] - c[1, :]])

    def test_2sc_many_axes(self, setup):
        # 2 orbitals per atom
        g = setup.g
        # functions for any-thing to SC
        idx = [[1], [2]]
        assert np.all(g.a2isc(idx) == [[[0, 0, 0]], [[-1, -1, 0]]])
        assert g.auc2sc(idx).shape == (2, g.n_s)
        idx = [[2], [4]]
        assert np.all(g.o2isc(idx) == [[[0, 0, 0]], [[-1, -1, 0]]])
        assert g.ouc2sc(idx).shape == (2, g.n_s)

    def test_reverse(self, setup):
        rev = setup.g.reverse()
        assert len(rev) == 2
        assert np.allclose(rev.xyz[::-1, :], setup.g.xyz)
        rev = setup.g.reverse(atoms=list(range(len(setup.g))))
        assert len(rev) == 2
        assert np.allclose(rev.xyz[::-1, :], setup.g.xyz)

    def test_scale1(self, setup):
        two = setup.g.scale(2)
        assert len(two) == len(setup.g)
        assert np.allclose(two.xyz[:, :] / 2.0, setup.g.xyz)

    def test_scale_vector_abc(self, setup):
        two = setup.g.scale([2, 1, 1], what="abc")
        assert len(two) == len(setup.g)
        # Check that cell has been scaled accordingly
        assert np.allclose(two.cell[0] / 2.0, setup.g.cell[0])
        assert np.allclose(two.cell[1:], setup.g.cell[1:])
        # Now check that fractional coordinates are still the same
        assert np.allclose(two.fxyz, setup.g.fxyz)

    def test_scale_vector_xyz(self, setup):
        two = setup.g.scale([2, 1, 1], what="xyz")
        assert len(two) == len(setup.g)
        # Check that cell has been scaled accordingly
        assert np.allclose(two.cell[:, 0] / 2.0, setup.g.cell[:, 0])
        assert np.allclose(two.cell[:, 1:], setup.g.cell[:, 1:])
        # Now check that fractional coordinates are still the same
        assert np.allclose(two.fxyz, setup.g.fxyz)

    def test_close1(self, setup):
        three = range(3)
        for ia in setup.mol:
            i = setup.mol.close(ia, R=(0.1, 1.1), atoms=three)
            if ia < 3:
                assert len(i[0]) == 1
            else:
                assert len(i[0]) == 0
            # Will only return results from [0,1,2]
            # but the fourth atom connects to
            # the third
            if ia in [0, 2, 3]:
                assert len(i[1]) == 1
            elif ia == 1:
                assert len(i[1]) == 2
            else:
                assert len(i[1]) == 0

    def test_close2(self, setup):
        mol = range(3, 5)
        for ia in setup.mol:
            i = setup.mol.close(ia, R=(0.1, 1.1), atoms=mol)
            assert len(i) == 2
        i = setup.mol.close([100, 100, 100], R=0.1)
        assert len(i) == 0
        i = setup.mol.close([100, 100, 100], R=0.1, ret_rij=True)
        for el in i:
            assert len(el) == 0
        i = setup.mol.close([100, 100, 100], R=0.1, ret_rij=True, ret_xyz=True)
        for el in i:
            assert len(el) == 0

    def test_close_arguments_2R(self, setup):
        g = setup.mol.copy()
        for args in ("isc", "xyz", "rij"):
            kwargs = {f"ret_{args}": True}
            ia, other = g.close(0, R=(0.1, 1.1), **kwargs)
            assert len(ia) == 2
            assert len(other) == 2
        ia, a, b, c = g.close(0, R=(0.1, 1.1), ret_isc=True, ret_rij=True, ret_xyz=True)
        assert len(ia) == 2
        assert len(a) == 2
        assert len(b) == 2
        assert len(c) == 2

    def test_close_arguments_1R(self, setup):
        g = setup.mol.copy()
        for args in ("isc", "xyz", "rij"):
            kwargs = {f"ret_{args}": True}
            ia, other = g.close(0, R=1.1, **kwargs)
            assert len(ia) == len(other)
        ia, a, b, c = g.close(0, R=1.1, ret_isc=True, ret_rij=True, ret_xyz=True)
        assert len(ia) == len(a)
        assert len(a) == len(b)
        assert len(b) == len(c)

    @pytest.mark.slow
    def test_close4(self, setup):
        # 2 * 200 ** 2
        g = setup.g * (200, 200, 1)
        i = g.close(0, R=(0.1, 1.43))
        assert len(i) == 2
        assert len(i[0]) == 1
        assert len(i[1]) == 3

    def test_close_within1(self, setup):
        three = range(3)
        for ia in setup.mol:
            shapes = [Sphere(0.1, setup.mol[ia]), Sphere(1.1, setup.mol[ia])]
            i = setup.mol.close(ia, R=(0.1, 1.1), atoms=three)
            ii = setup.mol.within(shapes, atoms=three)
            assert np.all(i[0] == ii[0])
            assert np.all(i[1] == ii[1])

    def test_close_within2(self, setup):
        g = setup.g.repeat(6, 0).repeat(6, 1)
        for ia in g:
            shapes = [Sphere(0.1, g[ia]), Sphere(1.5, g[ia])]
            i = g.close(ia, R=(0.1, 1.5))
            ii = g.within(shapes)
            assert np.all(i[0] == ii[0])
            assert np.all(i[1] == ii[1])

    def test_close_within3(self, setup):
        g = setup.g.repeat(6, 0).repeat(6, 1)
        args = {"ret_xyz": True, "ret_rij": True, "ret_isc": True}
        for ia in g:
            shapes = [Sphere(0.1, g[ia]), Sphere(1.5, g[ia])]
            i, xa, d, isc = g.close(ia, R=(0.1, 1.5), **args)
            ii, xai, di, isci = g.within(shapes, **args)
            for j in [0, 1]:
                assert np.all(i[j] == ii[j])
                assert np.allclose(xa[j], xai[j])
                assert np.allclose(d[j], di[j])
                assert np.allclose(isc[j], isci[j])

    def test_within_inf_small_translated(self, setup):
        g = setup.g.translate([0.05] * 3)
        lattice_3x3 = g.lattice.tile(3, 0).tile(3, 1)
        assert len(g.within_inf(lattice_3x3)[0]) == len(g) * 3**2

    def test_within_inf_nonperiodic(self, setup):
        g = setup.g.copy()

        # Even if the geometry has nsc > 1, if we set periodic=False
        # we should get only the atoms in the unit cell.
        g.set_nsc([3, 3, 1])

        ia, xyz, isc = g.within_inf(g.lattice, periodic=[False, False, False])

        assert len(ia) == g.na
        assert np.all(isc == 0)

        # Make sure that it also works with mixed periodic/non periodic directions
        ia, xyz, isc = g.within_inf(g.lattice, periodic=[True, False, False])

        assert len(ia) > g.na
        assert np.any(isc[:, 0] != 0)
        assert np.all(isc[:, 1] == 0)

    def test_within_inf_molecule(self, setup):
        g = setup.mol.translate([0.05] * 3)
        lattice = Lattice(1.5)
        for o in range(10):
            origin = [o - 0.5, -0.5, -0.5]
            lattice.origin = origin
            idx = g.within_inf(lattice)[0]
            assert len(idx) == 1
            assert idx[0] == o

    def test_within_inf_duplicates(self, setup):
        g = setup.g.copy()
        g.lattice.pbc = [True, True, False]
        lattice_3x3 = g.lattice.tile(3, 0).tile(3, 1)
        assert len(g.within_inf(lattice_3x3)[0]) == 25

    def test_within_inf_gh649(self):
        # see https://github.com/zerothi/sisl/issues/649

        # Create a geometry with an atom outside of the unit cell
        geometry = Geometry([-0.5, 0, 0], lattice=np.diag([2, 10, 10]))

        search = Lattice(np.diag([3, 10, 10]))
        ia, xyz, isc = geometry.within_inf(search, periodic=True)
        assert np.allclose(ia, 0)
        assert np.allclose(isc, [1, 0, 0])

        search = Lattice(np.diag([2, 10, 10]))
        ia, xyz, isc = geometry.within_inf(search, periodic=True)
        assert np.allclose(ia, 0)
        assert np.allclose(isc, [1, 0, 0])

    def test_close_sizes(self, setup):
        point = 0

        # Return index
        idx = setup.mol.close(point, R=0.1)
        assert len(idx) == 1
        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1))
        assert len(idx) == 2
        assert len(idx[0]) == 1
        assert not isinstance(idx[0], list)
        # Longer
        idx = setup.mol.close(point, R=(0.1, 1.1, 2.1))
        assert len(idx) == 3
        assert len(idx[0]) == 1

        # Return index
        idx = setup.mol.close(point, R=0.1, ret_xyz=True)
        assert len(idx) == 2
        assert len(idx[0]) == 1
        assert len(idx[1]) == 1
        assert idx[1].shape[0] == 1  # equivalent to above
        assert idx[1].shape[1] == 3

        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1), ret_xyz=True)
        # [[idx-1, idx-2], [coord-1, coord-2]]
        assert len(idx) == 2
        assert len(idx[0]) == 2
        assert len(idx[1]) == 2
        # idx-1
        assert len(idx[0][0].shape) == 1
        assert idx[0][0].shape[0] == 1
        # idx-2
        assert idx[0][1].shape[0] == 1
        # coord-1
        assert len(idx[1][0].shape) == 2
        assert idx[1][0].shape[1] == 3
        # coord-2
        assert idx[1][1].shape[1] == 3

        # Return index of two things
        idx = setup.mol.close(
            point, R=(0.1, 1.1), ret_xyz=True, ret_rij=True, ret_isc=True
        )
        # [[idx-1, idx-2], [coord-1, coord-2], [dist-1, dist-2], [isc-1, isc-2]]
        assert len(idx) == 4
        assert len(idx[0]) == 2
        assert len(idx[1]) == 2
        assert len(idx[2]) == 2
        assert len(idx[3]) == 2
        assert all(len(idx[0][0]) == len(t[0]) for t in idx)
        assert all(len(idx[0][1]) == len(t[1]) for t in idx)
        # idx-1
        assert len(idx[0][0].shape) == 1
        assert idx[0][0].shape[0] == 1
        # idx-2
        assert idx[0][1].shape[0] == 1
        # coord-1
        assert len(idx[1][0].shape) == 2
        assert idx[1][0].shape[1] == 3
        # coord-2
        assert idx[1][1].shape[1] == 3
        # dist-1
        assert len(idx[2][0].shape) == 1
        assert idx[2][0].shape[0] == 1
        # dist-2
        assert idx[2][1].shape[0] == 1
        # isc-1
        assert len(idx[3][0].shape) == 2
        assert idx[3][0].shape[1] == 3
        # isc-2
        assert idx[3][1].shape[1] == 3

        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1), ret_rij=True)
        # [[idx-1, idx-2], [dist-1, dist-2]]
        assert len(idx) == 2
        assert len(idx[0]) == 2
        assert len(idx[1]) == 2
        # idx-1
        assert len(idx[0][0].shape) == 1
        assert idx[0][0].shape[0] == 1
        # idx-2
        assert idx[0][1].shape[0] == 1
        # dist-1
        assert len(idx[1][0].shape) == 1
        assert idx[1][0].shape[0] == 1
        # dist-2
        assert idx[1][1].shape[0] == 1

    def test_close_sizes_none(self, setup):
        point = [100.0, 100.0, 100.0]

        # Return index
        idx = setup.mol.close(point, R=0.1)
        assert len(idx) == 0
        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1))
        assert len(idx) == 2
        assert len(idx[0]) == 0
        assert not isinstance(idx[0], list)
        # Longer
        idx = setup.mol.close(point, R=(0.1, 1.1, 2.1))
        assert len(idx) == 3
        assert len(idx[0]) == 0

        # Return index
        idx = setup.mol.close(point, R=0.1, ret_xyz=True)
        assert len(idx) == 2
        assert len(idx[0]) == 0
        assert len(idx[1]) == 0
        assert idx[1].shape[0] == 0  # equivalent to above
        assert idx[1].shape[1] == 3

        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1), ret_xyz=True)
        # [[idx-1, idx-2], [coord-1, coord-2]]
        assert len(idx) == 2
        assert len(idx[0]) == 2
        assert len(idx[1]) == 2
        # idx-1
        assert len(idx[0][0].shape) == 1
        assert idx[0][0].shape[0] == 0
        # idx-2
        assert idx[0][1].shape[0] == 0
        # coord-1
        assert len(idx[1][0].shape) == 2
        assert idx[1][0].shape[1] == 3
        # coord-2
        assert idx[1][1].shape[1] == 3

        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1), ret_xyz=True, ret_rij=True)
        # [[idx-1, idx-2], [coord-1, coord-2], [dist-1, dist-2]]
        assert len(idx) == 3
        assert len(idx[0]) == 2
        assert len(idx[1]) == 2
        # idx-1
        assert len(idx[0][0].shape) == 1
        assert idx[0][0].shape[0] == 0
        # idx-2
        assert idx[0][1].shape[0] == 0
        # coord-1
        assert len(idx[1][0].shape) == 2
        assert idx[1][0].shape[0] == 0
        assert idx[1][0].shape[1] == 3
        # coord-2
        assert idx[1][1].shape[0] == 0
        assert idx[1][1].shape[1] == 3
        # dist-1
        assert len(idx[2][0].shape) == 1
        assert idx[2][0].shape[0] == 0
        # dist-2
        assert idx[2][1].shape[0] == 0

        # Return index of two things
        idx = setup.mol.close(point, R=(0.1, 1.1), ret_rij=True)
        # [[idx-1, idx-2], [dist-1, dist-2]]
        assert len(idx) == 2
        assert len(idx[0]) == 2
        assert len(idx[1]) == 2
        # idx-1
        assert len(idx[0][0].shape) == 1
        assert idx[0][0].shape[0] == 0
        # idx-2
        assert idx[0][1].shape[0] == 0
        # dist-1
        assert len(idx[1][0].shape) == 1
        assert idx[1][0].shape[0] == 0
        # dist-2
        assert idx[1][1].shape[0] == 0

    def test_sparserij1(self, setup):
        rij = setup.g.sparserij()

    def test_bond_correct(self, setup):
        # Create ribbon
        rib = setup.g.tile(2, 1)
        # Convert the last atom to a H atom
        rib.atoms[-1] = Atom[1]
        ia = len(rib) - 1
        # Get bond-length
        idx, d = rib.close(ia, R=(0.1, 1000), ret_rij=True)
        i = np.argmin(d[1])
        d = d[1][i]
        rib.bond_correct(ia, idx[1][i])
        idx, d2 = rib.close(ia, R=(0.1, 1000), ret_rij=True)
        i = np.argmin(d2[1])
        d2 = d2[1][i]
        assert d != d2
        # Calculate actual radius
        assert d2 == (Atom[1].radius() + Atom[6].radius())

    def test_unit_cell_estimation1(self, setup):
        # Create new geometry with only the coordinates
        # and atoms
        geom = Geometry(setup.g.xyz, Atom[6])
        # Only check the two distances we know have sizes
        for i in range(2):
            # It cannot guess skewed axis
            assert not np.allclose(geom.cell[i, :], setup.g.cell[i, :])

    def test_unit_cell_estimation2(self, setup):
        # Create new geometry with only the coordinates
        # and atoms
        s1 = Lattice([2, 2, 2])
        g1 = Geometry([[0, 0, 0], [1, 1, 1]], lattice=s1)
        g2 = Geometry(np.copy(g1.xyz))
        assert np.allclose(g1.cell, g2.cell)

        # Assert that it correctly calculates the bond-length in the
        # directions of actual distance
        g1 = Geometry([[0, 0, 0], [1, 1, 0]], atoms="H", lattice=s1)
        g2 = Geometry(np.copy(g1.xyz))
        for i in range(2):
            assert np.allclose(g1.cell[i, :], g2.cell[i, :])
        assert not np.allclose(g1.cell[2, :], g2.cell[2, :])

    def test_distance1(self, setup):
        geom = Geometry(setup.g.xyz, Atom[6])
        # maxR is undefined
        with pytest.raises(ValueError):
            d = geom.distance()

    def test_distance2(self, setup):
        geom = Geometry(setup.g.xyz, Atom[6])
        with pytest.raises(ValueError):
            d = geom.distance(R=1.42, method="unknown_numpy_function")

    def test_distance3(self, setup):
        geom = setup.g.copy()
        d = geom.distance()
        assert len(d) == 1
        assert np.allclose(d, [1.42])

    def test_distance4(self, setup):
        geom = setup.g.copy()
        d = geom.distance(method=np.min)
        assert len(d) == 1
        assert np.allclose(d, [1.42])
        d = geom.distance(method=np.max)
        assert len(d) == 1
        assert np.allclose(d, [1.42])
        d = geom.distance(method="max")
        assert len(d) == 1
        assert np.allclose(d, [1.42])

    def test_distance5(self, setup):
        geom = setup.g.copy()
        d = geom.distance(R=np.inf)
        assert len(d) == 6
        d = geom.distance(0, R=1.42)
        assert len(d) == 1
        assert np.allclose(d, [1.42])

    def test_distance6(self, setup):
        # Create a 1D chain
        geom = Geometry([0] * 3, Atom(1, R=1.0), lattice=1)
        geom.set_nsc([77, 1, 1])
        d = geom.distance(0)
        assert len(d) == 1
        assert np.allclose(d, [1.0])

        # Do twice
        d = geom.distance(R=2)
        assert len(d) == 2
        assert np.allclose(d, [1.0, 2.0])

        # Do all
        d = geom.distance(R=np.inf)
        assert len(d) == 77 // 2
        # Add one due arange not adding the last item
        assert np.allclose(d, range(1, 78 // 2))

        # Create a 2D grid
        geom.set_nsc([3, 3, 1])
        d = geom.distance(R=2, atol=[0.4, 0.3, 0.2, 0.1])
        assert len(d) == 2  # 1, sqrt(2)
        # Add one due arange not adding the last item
        assert np.allclose(d, [1, 2**0.5])

        # Create a 2D grid
        geom.set_nsc([5, 5, 1])
        d = geom.distance(R=2, atol=[0.4, 0.3, 0.2, 0.1])
        assert len(d) == 3  # 1, sqrt(2), 2
        # Add one due arange not adding the last item
        assert np.allclose(d, [1, 2**0.5, 2])

    def test_distance7(self, setup):
        # Create a 1D chain
        geom = Geometry([0] * 3, Atom(1, R=1.0), lattice=1)
        geom.set_nsc([77, 1, 1])
        # Try with a short R and a long tolerance list
        # We know that the tolerance list prevails, because
        d = geom.distance(R=1, atol=np.ones(10) * 0.5)
        assert len(d) == 1
        assert np.allclose(d, [1.0])

    def test_distance8(self, setup):
        geom = Geometry([0] * 3, Atom(1, R=1.0), lattice=1)
        geom.set_nsc([77, 1, 1])
        d = geom.distance(0, method="min")
        assert len(d) == 1
        d = geom.distance(0, method="median")
        assert len(d) == 1
        d = geom.distance(0, method="mode")
        assert len(d) == 1

    def test_find_nsc1(self, setup):
        # Create a 1D chain
        geom = Geometry([0] * 3, Atom(1, R=1.0), lattice=1)
        geom.set_nsc([77, 77, 77])

        assert np.allclose(geom.find_nsc(), [3, 3, 3])
        assert np.allclose(geom.find_nsc(1), [77, 3, 77])
        assert np.allclose(geom.find_nsc([0, 2]), [3, 77, 3])
        assert np.allclose(geom.find_nsc([0, 2], R=2.00000001), [5, 77, 5])

        geom.set_nsc([1, 1, 1])
        assert np.allclose(geom.find_nsc([0, 2], R=2.00000001), [5, 1, 5])

        geom.set_nsc([5, 1, 5])
        assert np.allclose(geom.find_nsc([0, 2], R=0.9999), [1, 1, 1])

    def test_find_nsc2(self, setup):
        # 2 ** 0.5 ensures lattice vectors with length 1
        geom = sisl_geom.fcc(2**0.5, Atom(1, R=1.0001))
        geom.set_nsc([77, 77, 77])

        assert np.allclose(geom.find_nsc(), [3, 3, 3])
        assert np.allclose(geom.find_nsc(1), [77, 3, 77])
        assert np.allclose(geom.find_nsc([0, 2]), [3, 77, 3])
        assert np.allclose(geom.find_nsc([0, 2], R=2.00000001), [5, 77, 5])

        geom.set_nsc([1, 1, 1])
        assert np.allclose(geom.find_nsc([0, 2], R=2.00000001), [5, 1, 5])

        geom.set_nsc([5, 1, 5])
        assert np.allclose(geom.find_nsc([0, 2], R=0.9999), [1, 1, 1])

    def test_argumentparser1(self, setup):
        setup.g.ArgumentParser()
        setup.g.ArgumentParser(**setup.g._ArgumentParser_args_single())

    def test_argumentparser2(self, setup, **kwargs):
        p, ns = setup.g.ArgumentParser(**kwargs)

        # Try all options
        opts = [
            "--origin",
            "--center-of",
            "mass",
            "--center-of",
            "xyz",
            "--center-of",
            "position",
            "--center-of",
            "cell",
            "--unit-cell",
            "translate",
            "--unit-cell",
            "mod",
            "--rotate",
            "90",
            "x",
            "--rotate",
            "90",
            "y",
            "--rotate",
            "90",
            "z",
            "--add",
            "0,0,0",
            "6",
            "--swap",
            "0",
            "1",
            "--repeat",
            "2",
            "x",
            "--repeat",
            "2",
            "y",
            "--repeat",
            "2",
            "z",
            "--tile",
            "2",
            "x",
            "--tile",
            "2",
            "y",
            "--tile",
            "2",
            "z",
            "--untile",
            "2",
            "z",
            "--untile",
            "2",
            "y",
            "--untile",
            "2",
            "x",
        ]
        if kwargs.get("limit_arguments", True):
            opts.extend(
                ["--rotate", "-90", "x", "--rotate", "-90", "y", "--rotate", "-90", "z"]
            )
        else:
            opts.extend(
                [
                    "--rotate-x",
                    " -90",
                    "--rotate-y",
                    " -90",
                    "--rotate-z",
                    " -90",
                    "--repeat-x",
                    "2",
                    "--repeat-y",
                    "2",
                    "--repeat-z",
                    "2",
                ]
            )

        args = p.parse_args(opts, namespace=ns)

        if len(kwargs) == 0:
            self.test_argumentparser2(setup, **setup.g._ArgumentParser_args_single())

    def test_set_supercell(self, setup):
        # check for deprecation
        s1 = Lattice([2, 2, 2])
        g1 = Geometry([[0, 0, 0], [1, 1, 1]], lattice=[2, 2, 1])
        with pytest.warns(SislDeprecation) as deps:
            g1.set_supercell(s1)
        assert g1.lattice == s1
        assert len(deps) == 1

    def test_attach1(self, setup):
        g = setup.g.attach(0, setup.mol, 0, dist=1.42, axis=2)
        g = setup.g.attach(0, setup.mol, 0, dist="calc", axis=2)
        g = setup.g.attach(0, setup.mol, 0, dist=[0, 0, 1.42])

    def test_mirror_function(self, setup):
        g = setup.g
        for plane in ["xy", "xz", "yz", "ab", "bc", "ac"]:
            g.mirror(plane)

        assert g.mirror("xy") == g.mirror("z")
        assert g.mirror("xy") == g.mirror([0, 0, 1])

        assert g.mirror("xy", [0]) == g.mirror([0, 0, 1], [0])

    def test_mirror_point(self):
        g = Geometry([[0, 0, 0], [0, 0, 1]])
        out = g.mirror("z")
        assert np.allclose(out.xyz[:, 2], [0, -1])
        assert np.allclose(out.xyz[:, :2], 0)
        out = g.mirror("z", point=(0, 0, 0.5))
        assert np.allclose(out.xyz[:, 2], [1, 0])
        assert np.allclose(out.xyz[:, :2], 0)
        out = g.mirror("z", point=(0, 0, 1))
        assert np.allclose(out.xyz[:, 2], [2, 1])
        assert np.allclose(out.xyz[:, :2], 0)

    def test_pickle(self, setup):
        import pickle as p

        s = p.dumps(setup.g)
        n = p.loads(s)
        assert n == setup.g

    def test_geometry_names(self):
        g = sisl_geom.graphene()

        assert len(g.names) == 0
        g["A"] = 1
        assert len(g.names) == 1
        g[[1, 2]] = "B"
        assert len(g.names) == 2
        g.names.delete_name("B")
        assert len(g.names) == 1

        # Add new group
        g["B"] = [0, 2]

        for name in g.names:
            assert name in ["A", "B"]

        str(g)

        assert np.allclose(g["B"], g[[0, 2], :])
        assert np.allclose(g.axyz("B"), g[[0, 2], :])

        del g.names["B"]
        assert len(g.names) == 1

    def test_geometry_groups_raise(self):
        g = sisl_geom.graphene()
        g["A"] = 1
        with pytest.raises(SislError):
            g["A"] = [1, 2]

    def test_geometry_as_primary_raise_nondivisable(self):
        g = sisl_geom.graphene()
        with pytest.raises(ValueError):
            g.as_primary(3)

    def test_geometry_untile_raise_nondivisable(self):
        g = sisl_geom.graphene()
        with pytest.raises(ValueError):
            g.untile(3, 0)

    def test_geometry_iR_negative_R(self):
        g = sisl_geom.graphene()
        with pytest.raises(ValueError):
            g.iR(R=-1.0)

    @pytest.mark.parametrize(
        "geometry",
        [
            sisl_geom.graphene(),
            sisl_geom.diamond(),
            sisl_geom.sc(1.4, Atom[1]),
            sisl_geom.fcc(1.4, Atom[1]),
            sisl_geom.bcc(1.4, Atom[1]),
            sisl_geom.hcp(1.4, Atom[1]),
        ],
    )
    def test_geometry_as_primary(self, geometry):
        prod = itertools.product
        x_reps = [1, 4, 3]
        y_reps = [1, 4, 5]
        z_reps = [1, 4, 6]
        tile_rep = ["r", "t"]

        na_primary = len(geometry)
        for x, y, z in prod(x_reps, y_reps, z_reps):
            if x == y == z == 1:
                continue
            for a, b, c in prod(tile_rep, tile_rep, tile_rep):
                G = ((geometry * ([x, 1, 1], a)) * ([1, y, 1], b)) * ([1, 1, z], c)
                p = G.as_primary(na_primary)
                assert np.allclose(p.xyz, geometry.xyz)
                assert np.allclose(p.cell, geometry.cell)

    def test_geometry_as_primary_without_super(self):
        g = sisl_geom.graphene()
        p = g.as_primary(len(g))
        assert g == p

        G = g.tile(2, 0).tile(3, 1)
        p, supercell = G.as_primary(len(g), ret_super=True)
        assert np.allclose(p.xyz, g.xyz)
        assert np.allclose(p.cell, g.cell)
        assert np.all(supercell == [2, 3, 1])

    # Test ASE (but only fail if present)

    def test_geometry_dispatch(self):
        pytest.importorskip("ase", reason="ase not available")
        gr = sisl_geom.graphene()
        to_ase = gr.to.ase()

        ase_rotate = si.rotate(to_ase, 30, [0, 0, 1])
        assert isinstance(ase_rotate, type(to_ase))
        ase_sisl_rotate = si.rotate(to_ase, 30, [0, 0, 1], ret_sisl=True)
        assert isinstance(ase_sisl_rotate, Geometry)
        geom_rotate = si.rotate(gr, 30, [0, 0, 1])

        assert geom_rotate.equal(ase_sisl_rotate, R=False)

    def test_geometry_ase_new_to(self):
        pytest.importorskip("ase", reason="ase not available")
        gr = sisl_geom.graphene()
        to_ase = gr.to.ase()
        from_ase = gr.new(to_ase)
        assert gr.equal(from_ase, R=False)

    def test_geometry_ase_run_center(self):
        pytest.importorskip("ase", reason="ase not available")
        gr = sisl_geom.graphene()
        ase_atoms = gr.to.ase()
        from_ase = si.center(ase_atoms)
        assert np.allclose(gr.center(), from_ase)

    @pytest.mark.xfail(
        reason="pymatgen backconversion sets nsc=[3, 3, 3], we need to figure this out"
    )
    def test_geometry_pymatgen_to(self):
        pytest.importorskip("pymatgen", reason="pymatgen not available")
        gr = sisl_geom.graphene()
        to_pymatgen = gr.to.pymatgen()
        from_pymatgen = gr.new(to_pymatgen)
        assert gr.equal(from_pymatgen, R=False)
        # TODO add test for Molecule as well (cell will then be different)

    def test_geometry_pandas_to(self):
        pytest.importorskip("pandas", reason="pandas not available")
        gr = sisl_geom.graphene()
        df = gr.to.dataframe()
        assert np.allclose(df["x"], gr.xyz[:, 0])
        assert np.allclose(df["y"], gr.xyz[:, 1])
        assert np.allclose(df["z"], gr.xyz[:, 2])

    def test_geometry_overlapping_atoms(self):
        gr22 = sisl_geom.graphene().tile(2, 0).tile(2, 1)
        gr44 = gr22.tile(2, 0).tile(2, 1)
        offset = np.array([0.2, 0.4, 0.4])
        gr22 = gr22.translate(offset)
        idx = np.arange(gr22.na)
        np.random.shuffle(idx)
        gr22 = gr22.sub(idx)
        idx22, idx44 = gr22.overlap(gr44, offset=-offset)
        assert np.allclose(idx22, np.arange(gr22.na))
        assert np.allclose(idx44, idx)


def test_geometry_sort_simple():
    bi = sisl_geom.bilayer().tile(2, 0).repeat(3, 1)

    # the default tolerance is 1e-9, and since we are sorting
    # in each group, we may in the end find another ordering
    atol = 1e-9

    for i in [0, 1, 2]:
        s = bi.sort(axes=i)
        assert np.all(np.diff(s.xyz[:, i]) >= -atol)
        s = bi.sort(lattice=i)
        assert np.all(np.diff(s.fxyz[:, i] * bi.lattice.length[i]) >= -atol)

    s, idx = bi.sort(axes=0, lattice=1, ret_atoms=True)
    assert np.all(np.diff(s.xyz[:, 0]) >= -atol)
    for ix in idx:
        assert np.all(np.diff(bi.fxyz[ix, 1]) >= -atol)

    s, idx = bi.sort(
        axes=0, ascending=False, lattice=1, vector=[0, 0, 1], ret_atoms=True
    )
    assert np.all(np.diff(s.xyz[:, 0]) >= -atol)
    for ix in idx:
        # idx is according to bi
        assert np.all(np.diff(bi.fxyz[ix, 1] * bi.lattice.length[i]) <= atol)


def test_geometry_sort_int():
    bi = sisl_geom.bilayer().tile(2, 0).repeat(3, 1)

    # the default tolerance is 1e-9, and since we are sorting
    # in each group, we may in the end find another ordering
    atol = 1e-9

    for i in [0, 1, 2]:
        s = bi.sort(axes0=i)
        assert np.all(np.diff(s.xyz[:, i]) >= -atol)
        s = bi.sort(lattice3=i)
        assert np.all(np.diff(s.fxyz[:, i] * bi.lattice.length[i]) >= -atol)

    s, idx = bi.sort(axes12314=0, lattice0=1, ret_atoms=True)
    assert np.all(np.diff(s.xyz[:, 0]) >= -atol)
    for ix in idx:
        assert np.all(np.diff(bi.fxyz[ix, 1]) >= -atol)

    s, idx = bi.sort(
        ascending1=True, axes15=0, ascending0=False, lattice235=1, ret_atoms=True
    )
    assert np.all(np.diff(s.xyz[:, 0]) >= -atol)
    for ix in idx:
        # idx is according to bi
        assert np.all(np.diff(bi.fxyz[ix, 1] * bi.lattice.length[i]) <= atol)


def test_geometry_ellipsis():
    gr = sisl_geom.graphene()
    assert np.allclose(gr.axyz(...), gr.axyz(None))


def test_geometry_sort_atom():
    bi = sisl_geom.bilayer().tile(2, 0).repeat(2, 1)

    atom = [[2, 0], [3, 1]]
    out, atom = bi.sort(atoms=atom, ret_atoms=True)

    atom = np.concatenate(atom)
    all_atoms = np.arange(len(bi))
    all_atoms[np.sort(atom)] = atom[:]

    assert np.allclose(out.xyz, bi.sub(all_atoms).xyz)


def test_geometry_sort_func():
    bi = sisl_geom.bilayer().tile(2, 0).repeat(2, 1)

    def reverse(geometry, atoms, **kwargs):
        return atoms[::-1]

    atoms = [[2, 0], [3, 1]]
    out = bi.sort(func=reverse, atoms=atoms)

    all_atoms = np.arange(len(bi))
    all_atoms[1] = 2
    all_atoms[2] = 1

    assert np.allclose(out.xyz, bi.sub(all_atoms).xyz)

    bi_again = bi.sort(func=(reverse, reverse), atoms=atoms)

    # Ensure that they are swapped
    atoms = [2, 0]
    out = bi.sort(func=reverse, atoms=atoms)

    assert np.allclose(out.xyz, bi.xyz)

    out = bi.sort(func=reverse)
    all_atoms = np.arange(len(bi))[::-1]
    assert np.allclose(out.xyz, bi.sub(all_atoms).xyz)


def test_geometry_sort_func_sort():
    bi = sisl_geom.bilayer().tile(2, 0).repeat(2, 1)

    # Sort according to another cell fractional coordinates
    fcc = sisl_geom.fcc(2.4, Atom(6))

    def fcc_fracs(axis):
        def _(geometry):
            return np.dot(geometry.xyz, fcc.icell.T)[:, axis]

        return _

    out = bi.sort(func_sort=(fcc_fracs(0), fcc_fracs(2)))


def test_geometry_sort_group():
    bi = (
        sisl_geom.bilayer(bottom_atoms=Atom[6], top_atoms=(Atom[5], Atom[7]))
        .tile(2, 0)
        .repeat(2, 1)
    )

    out = bi.sort(group="Z")

    assert np.allclose(out.atoms.Z[:4], 5)
    assert np.allclose(out.atoms.Z[4:12], 6)
    assert np.allclose(out.atoms.Z[12:16], 7)

    out = bi.sort(group=("symbol", "C", None))

    assert np.allclose(out.atoms.Z[:8], 6)

    C = bi.sort(group=("symbol", "C", None))
    BN = bi.sort(group=("symbol", None, "C"))
    BN2 = bi.sort(group=("symbol", ["B", "N"], "C"))
    # For these simple geometries symbol and tag are the same
    BN3 = bi.sort(group=("tag", ["B", "N"], "C"))

    # none of these atoms should be the same
    assert not np.any(np.isclose(C.atoms.Z, BN.atoms.Z))
    # All these sorting algorithms are the same
    assert np.allclose(BN.atoms.Z, BN2.atoms.Z)
    assert np.allclose(BN.atoms.Z, BN3.atoms.Z)

    mass = bi.sort(group="mass")
    Z = bi.sort(group="Z")
    assert np.allclose(mass.atoms.Z, Z.atoms.Z)


def test_geometry_sort_fail_keyword():
    with pytest.raises(ValueError):
        sisl_geom.bilayer().sort(not_found_keyword=True)


@pytest.mark.category
@pytest.mark.geom_category
def test_geometry_sanitize_atom_category():
    bi = (
        sisl_geom.bilayer(bottom_atoms=Atom[6], top_atoms=(Atom[5], Atom[7]))
        .tile(2, 0)
        .repeat(2, 1)
    )
    C_idx = (bi.atoms.Z == 6).nonzero()[0]
    check_C = bi.axyz(C_idx)
    only_C = bi.axyz(Atom[6])
    assert np.allclose(only_C, check_C)
    only_C = bi.axyz(bi.atoms.Z == 6)
    assert np.allclose(only_C, check_C)
    # with dict redirect
    only_C = bi.axyz({"Z": 6})
    assert np.allclose(only_C, check_C)
    # Using a dict that has multiple keys. This basically checks
    # that it accepts generic categories such as the AndCategory
    bi2 = bi.copy()
    bi2.atoms["C"] = Atom("C", R=1.9)
    only_C = bi2.axyz({"Z": 6, "neighbors": 3})
    assert np.allclose(only_C, check_C)

    tup_01 = (0, 2)
    list_01 = [0, 2]
    ndarray_01 = np.array(list_01)
    assert np.allclose(bi._sanitize_atoms(tup_01), bi._sanitize_atoms(list_01))
    assert np.allclose(bi._sanitize_atoms(ndarray_01), bi._sanitize_atoms(list_01))


def test_geometry_sanitize_atom_shape():
    bi = (
        sisl_geom.bilayer(bottom_atoms=Atom[6], top_atoms=(Atom[5], Atom[7]))
        .tile(2, 0)
        .repeat(2, 1)
    )
    cube = Cube(10)
    assert len(bi.axyz(cube)) != 0


def test_geometry_sanitize_atom_0_length():
    gr = sisl_geom.graphene()
    assert len(gr.axyz([])) == 0


@pytest.mark.parametrize("atoms", [[True, False], (True, False), [0], (0,)])
def test_geometry_sanitize_atom_other_bool(atoms):
    gr = sisl_geom.graphene()
    assert len(gr.axyz(atoms)) == 1


def test_geometry_sanitize_atom_0_length_float_fail():
    gr = sisl_geom.graphene()
    with pytest.raises(IndexError):
        # it raises an error because np.float64 is used
        gr.axyz(np.array([1], dtype=np.float64))


def test_geometry_sanitize_atom_bool():
    gr = sisl_geom.graphene()
    assert gr.axyz(True).shape == (gr.na, 3)
    assert gr.axyz(False).shape == (0, 3)


def test_geometry_sanitize_orbs():
    bot = Atom(6, [1, 2, 3])
    top = [Atom(5, [1, 2]), Atom(7, [1, 2])]

    bi = sisl_geom.bilayer(bottom_atoms=bot, top_atoms=top).tile(2, 0).repeat(2, 1)

    C_idx = (bi.atoms.Z == 6).nonzero()[0]
    assert np.allclose(bi._sanitize_orbs({bot: [0]}), bi.firsto[C_idx])
    assert np.allclose(bi._sanitize_orbs({bot: 1}), bi.firsto[C_idx] + 1)
    assert np.allclose(
        bi._sanitize_orbs({bot: [1, 2]}), np.add.outer(bi.firsto[C_idx], [1, 2]).ravel()
    )


def test_geometry_sub_orbitals():
    atom = Atom(6, [1, 2, 3])
    gr = sisl_geom.graphene(atoms=atom)
    assert gr.no == 6

    # try and do sub
    gr2 = gr.sub_orbital(atom, [atom.orbitals[0], atom.orbitals[2]])
    assert gr2.atoms[0][0] == atom.orbitals[0]
    assert gr2.atoms[0][1] == atom.orbitals[2]

    gr2 = gr.sub_orbital(atom, atom.orbitals[1])
    assert gr2.atoms[0][0] == atom.orbitals[1]


def test_geometry_new_xyz(sisl_tmp):
    # test that Geometry.new works
    out = sisl_tmp("out.xyz")
    C = Atom[6]
    gr = sisl_geom.graphene(atoms=C)
    # writing doesn't save orbital information, so we force
    # an empty atom
    gr.write(out)

    gr2 = Geometry.new(out)
    assert np.allclose(gr.xyz, gr2.xyz)
    assert gr == gr2

    gr2 = gr.new(out)
    assert np.allclose(gr.xyz, gr2.xyz)
    assert gr == gr2


def test_translate2uc():
    gr = sisl_geom.graphene() * (2, 3, 1)
    gr = gr.move([5, 5, 5])
    gr2 = gr.translate2uc()
    assert not np.allclose(gr.xyz, gr2.xyz)


def test_translate2uc_axes():
    gr = sisl_geom.graphene() * (2, 3, 1)
    gr = gr.move([5, 5, 5])
    gr_once = gr.translate2uc()
    gr_individual = gr.translate2uc(axes=0)
    assert not np.allclose(gr_once.xyz, gr_individual.xyz)
    gr_individual = gr_individual.translate2uc(axes=1)
    assert np.allclose(gr_once.xyz, gr_individual.xyz)
    gr_individual = gr_individual.translate2uc(axes=2)
    assert np.allclose(gr_once.xyz, gr_individual.xyz)


def test_fortran_contiguous():
    # for #748
    xyz = np.zeros([10, 3], order="F")
    geom = Geometry(xyz)
    assert geom.xyz.flags.c_contiguous


def test_as_supercell_graphene():
    gr = sisl_geom.graphene()
    grsc = gr.as_supercell()
    assert np.allclose(grsc.xyz[: len(gr)], gr.xyz)
    assert np.allclose(grsc.axyz(np.arange(gr.na_s)), gr.axyz(np.arange(gr.na_s)))


def test_as_supercell_fcc():
    g = sisl_geom.fcc(2**0.5, Atom(1, R=1.0001))
    gsc = g.as_supercell()
    assert np.allclose(gsc.xyz[: len(g)], g.xyz)
    assert np.allclose(gsc.axyz(np.arange(g.na_s)), g.axyz(np.arange(g.na_s)))


def test_new_inherit_class():
    g = sisl_geom.fcc(2**0.5, Atom(1, R=1.0001))

    class MyClass(Geometry):
        pass

    g2 = MyClass.new(g)
    assert g2.__class__ == MyClass
    assert g2 == g
    g2 = Geometry.new(g)
    assert g2.__class__ == Geometry
    assert g2 == g


def test_sc_warn():
    with pytest.warns(SislDeprecation):
        lattice = sisl_geom.graphene().sc


def test_geometry_apply():
    from functools import partial

    g = sisl_geom.fcc(2**0.5, Atom(1, R=(1.0001, 2)))
    data = np.random.rand(3, g.no, 4)
    data_dup = g.apply(data, "sum", lambda x: x, segments="orbitals", axis=1)
    assert data_dup.shape == data.shape
    assert np.allclose(data, data_dup)

    data_atom = g.apply(data, "sum", partial(g.a2o, all=True), axis=1)
    assert data_atom.shape == (3, g.na, 4)
    assert np.allclose(data_atom[:, 0], data[:, :2].sum(1))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_tshs_si_pdos_kgrid(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    assert si.version == 1
    HS1 = si.read_hamiltonian()
    f = sisl_tmp("tmp.TSHS")
    HS1.write(f)
    si = sisl.get_sile(f)
    HS2 = si.read_hamiltonian()
    assert HS1._csr.spsame(HS2._csr)
    HS1.finalize()
    HS2.finalize()
    assert np.allclose(HS1._csr._D, HS2._csr._D)


@pytest.mark.filterwarnings("ignore", message="*Casting complex values")
def test_tshs_si_pdos_dtypes_eigs(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    data = []
    eigs = None
    k = [0.1] * 3
    for dtype in (np.float32, np.float64, np.complex64, np.complex128):
        HS = si.read_hamiltonian(dtype=dtype)
        data.append(HS)
        assert HS.dtype == dtype

        if eigs is None:
            eigs = HS.eigh(k)
        else:
            assert np.allclose(eigs, HS.eigh(k), atol=1e-5)

    f = sisl_tmp("tmp.TSHS")
    fnc = sisl_tmp("tmp.nc")
    for HS in data:
        HS.write(f)
        HS1 = HS.read(f)
        assert np.allclose(eigs, HS1.eigh(k), atol=1e-5)

        HS.write(fnc)
        HS1 = HS.read(fnc)
        assert np.allclose(eigs, HS1.eigh(k), atol=1e-5)


def test_tshs_si_pdos_kgrid_tofromnc(sisl_files, sisl_tmp):
    pytest.importorskip("netCDF4")
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    HS1 = si.read_hamiltonian()
    f = sisl_tmp("tmp.TSHS")
    fnc = sisl_tmp("tmp.nc")

    HS1.write(f)
    HS1.write(fnc)

    HS2 = sisl.get_sile(f).read_hamiltonian()
    HS2nc = sisl.get_sile(fnc).read_hamiltonian()
    assert HS1._csr.spsame(HS2._csr)
    assert HS1._csr.spsame(HS2nc._csr)
    HS1.finalize()
    HS2.finalize()
    assert np.allclose(HS1._csr._D, HS2._csr._D)
    HS2nc.finalize()
    assert np.allclose(HS1._csr._D, HS2nc._csr._D)


def test_tshs_si_pdos_kgrid_repeat_tile(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    HS = si.read_hamiltonian()
    HSr = HS.repeat(3, 2).repeat(3, 0).repeat(3, 1)
    HSt = HS.tile(3, 2).tile(3, 0).tile(3, 1)
    assert np.allclose(HSr.eigh(), HSt.eigh())


def test_tshs_si_pdos_kgrid_repeat_tile_not_used(sisl_files, sisl_tmp):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    HS = si.read_hamiltonian()
    for i in range(HS.no):
        HS._csr._extend_empty(i, 3 + i % 3)
    HSt = HS.tile(3, 2).tile(3, 0).tile(3, 1)
    HSr = HS.repeat(3, 2).repeat(3, 0).repeat(3, 1)
    assert np.allclose(HSr.eigh(), HSt.eigh())


def test_tshs_soc_pt2_xx(sisl_files, sisl_tmp):
    fdf = sisl.get_sile(sisl_files("siesta", "Pt2_soc", "Pt2.fdf"))
    HS1 = fdf.read_hamiltonian()
    f = sisl_tmp("tmp.TSHS")
    HS1.write(f)
    si = sisl.get_sile(f)
    HS2 = si.read_hamiltonian()
    assert HS1._csr.spsame(HS2._csr)
    HS1.finalize()
    HS2.finalize()
    assert np.allclose(HS1._csr._D, HS2._csr._D)


@pytest.mark.filterwarnings("ignore", message="*Casting complex values")
def test_tshs_soc_pt2_xx_dtypes(sisl_files, sisl_tmp):
    fdf = sisl.get_sile(sisl_files("siesta", "Pt2_soc", "Pt2.fdf"))
    data = []
    eigs = None
    k = [0.1] * 3
    for dtype in (np.float32, np.float64, np.complex64, np.complex128):
        HS = fdf.read_hamiltonian(dtype=dtype)
        data.append(HS)
        assert HS.dtype == dtype

        if eigs is None:
            eigs = HS.eigh(k)
        else:
            assert np.allclose(eigs, HS.eigh(k), atol=1e-5)

    f = sisl_tmp("tmp.TSHS")
    fnc = sisl_tmp("tmp.nc")
    for HS in data:
        HS.write(f)
        HS1 = HS.read(f)
        assert np.allclose(eigs, HS1.eigh(k), atol=1e-5)

        HS.write(fnc)
        HS1 = HS.read(fnc)
        assert np.allclose(eigs, HS1.eigh(k), atol=1e-5)


def test_tshs_soc_pt2_xx_pdos(sisl_files):
    fdf = sisl.get_sile(sisl_files("siesta", "Pt2_soc", "Pt2.fdf"))
    sc = fdf.read_lattice(order="TSHS")
    HS = fdf.read_hamiltonian()
    assert np.allclose(sc.cell, HS.geometry.lattice.cell)
    HS.eigenstate().PDOS(np.linspace(-2, 2, 400))


def test_tshs_warn(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))

    # check number of orbitals
    geom = si.read_geometry()
    geom._atoms = sisl.Atoms([sisl.Atom(i + 1) for i in range(geom.na)])
    with pytest.warns(sisl.SislWarning, match="number of orbitals"):
        si.read_hamiltonian(geometry=geom)

    # check cell
    geom = si.read_geometry()
    geom.lattice.cell[:, :] += 1.0
    with pytest.warns(sisl.SislWarning, match="lattice vectors"):
        si.read_hamiltonian(geometry=geom)

    # check atomic coordinates
    geom = si.read_geometry()
    geom.xyz[0, :] += 10.0
    with pytest.warns(sisl.SislWarning, match="atomic coordinates"):
        si.read_hamiltonian(geometry=geom)

    # check supercell
    geom = si.read_geometry()
    geom.set_nsc([1, 1, 1])
    with pytest.warns(sisl.SislWarning, match="supercell"):
        si.read_hamiltonian(geometry=geom)


def test_tshs_error(sisl_files):
    # reading with a wrong geometry
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))

    # check number of orbitals
    geom = si.read_geometry()
    geom = sisl.Geometry(np.random.rand(geom.na + 1, 3))
    with pytest.raises(sisl.SileError):
        si.read_hamiltonian(geometry=geom)


def test_tshs_si_pdos_kgrid_overlap(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.TSHS"))
    HS = si.read_hamiltonian()
    S = si.read_overlap()
    assert HS._csr.spsame(S._csr)
    HS.finalize()
    S.finalize()
    assert np.allclose(HS._csr._D[:, HS.S_idx], S._csr._D[:, 0])


@pytest.mark.filterwarnings("ignore", message="*is NOT Hermitian for on-site")
def test_tshs_spin_orbit_tshs2nc2tshs(sisl_tmp):
    pytest.importorskip("netCDF4")
    H1 = sisl.Hamiltonian(sisl.geom.graphene(), spin=sisl.Spin("SO"))
    H1.construct(
        (
            [0.1, 1.44],
            [
                [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8],
                [0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9],
            ],
        )
    )

    fdf_file = sisl_tmp("RUN.fdf")
    f1 = sisl_tmp("tmp1.TSHS")
    f2 = sisl_tmp("tmp1.nc")
    H1.write(f1)
    H1.finalize()
    H2 = sisl.get_sile(f1).read_hamiltonian()
    H2.write(f2)
    H3 = sisl.get_sile(f2).read_hamiltonian()
    open(fdf_file, "w").writelines(["SystemLabel tmp1"])
    fdf = sisl.get_sile(fdf_file)
    assert np.allclose(
        fdf.read_lattice(order="nc").cell, fdf.read_lattice(order="TSHS").cell
    )
    assert H1._csr.spsame(H2._csr)
    assert np.allclose(H1._csr._D, H2._csr._D)
    assert H1._csr.spsame(H3._csr)
    assert np.allclose(H1._csr._D, H3._csr._D)


def test_tshs_missing_diagonal(sisl_tmp):
    H1 = sisl.Hamiltonian(sisl.geom.graphene())
    H1.construct(([0.1, 1.44], [0.0, -2.7]))
    # remove diagonal component here
    del H1[0, 0]

    f1 = sisl_tmp("tmp1.TSHS")
    with pytest.warns(sisl.SislWarning, match=r"changes the sparsity pattern"):
        H1.write(f1)

    f2 = sisl_tmp("tmp2.TSHS")
    H2 = sisl.get_sile(f1).read_hamiltonian()
    H2.write(f2)
    H3 = sisl.get_sile(f2).read_hamiltonian()

    H1.finalize()
    assert not H1._csr.spsame(H2._csr)
    assert H2._csr.spsame(H3._csr)
    assert np.allclose(H2._csr._D, H3._csr._D)
    H1[0, 0] = 0.0
    H1.finalize()
    assert H1._csr.spsame(H2._csr)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.siesta.times import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_h2o_dipole_times(sisl_files):
    f1 = sisl_files("siesta", "H2O_dipole", "TIMES")
    f2 = sisl_files("siesta", "H2O_dipole", "h2o_dipole.times")
    t1 = timesSileSiesta(f1).read_data()
    t2 = timesSileSiesta(f2).read_data()

    assert np.allclose(t1.to_array(), t2.to_array())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Coefficient, State, StateC, geom

pytestmark = [pytest.mark.physics, pytest.mark.state]


def ar(*args):
    l = np.prod(args)
    return np.arange(l, dtype=np.float64).reshape(*args)


def ortho_matrix(n, m=None):
    if m is None:
        m = n
    max_nm = max(n, m)
    from scipy.linalg import qr

    H = np.random.randn(max_nm, max_nm) + 1j * np.random.randn(max_nm, max_nm)
    Q, _ = qr(H)
    M = Q.dot(np.conjugate(Q.T))
    return M[:n, :]


def outer(v):
    return np.outer(v, np.conjugate(v))


def couter(c, v):
    return np.outer(v * c, np.conjugate(v))


def test_coefficient_creation_simple():
    c = Coefficient(ar(6))
    str(c)
    assert len(c) == 6
    assert c.shape == (6,)
    assert c.dtype == np.float64
    assert c.dkind == "f"
    assert len(c.sub(1)) == 1
    assert np.allclose(c.sub(1).c, 1)
    assert len(c.sub([1, 4])) == 2
    assert np.allclose(c.sub([1, 4]).c, [1, 4])
    assert np.allclose(c[1, 4].c, [1, 4])


def test_coefficient_creation_info():
    c = Coefficient(ar(6), geom.graphene(), k="HELLO")
    assert np.allclose(c.parent.xyz, geom.graphene().xyz)
    assert c.info["k"] == "HELLO"


def test_coefficient_copy():
    c = Coefficient(ar(6), geom.graphene(), k="HELLO", test="test")
    cc = c.copy()
    assert cc.info["k"] == "HELLO"
    assert cc.info["test"] == "test"


def test_coefficient_sub():
    state = ar(10)
    state = Coefficient(state)
    assert len(state) == 10
    for i in range(len(state)):
        assert len(state.sub(i)) == 1
    for i, sub in enumerate(state):
        assert len(sub) == 1

    assert np.allclose(
        state.sub(np.array([False, True, False, True])).c, state.sub([1, 3]).c
    )

    sub = state.sub(np.array([False, True, False, True]))
    state.sub([1, 3], inplace=True)
    assert np.allclose(sub.c, state.c)


def test_coefficient_iter():
    c = Coefficient(ar(6))
    i = 0
    for C in c:
        assert len(C) == 1
        i += 1
    assert i == 6
    for i, C in enumerate(c.iter(True)):
        assert C == c.c[i]


def test_state_creation():
    state = State(ar(6))
    assert len(state) == 1
    assert state.shape == (1, 6)
    assert state.norm2()[0] == pytest.approx((ar(6) ** 2).sum())
    state_c = state.copy()
    assert len(state) == len(state_c)
    str(state)


def test_state_repr():
    state = State(ar(6))
    str(state)
    state = State(ar(6), parent=geom.graphene())
    str(state)


def test_state_dkind():
    state = State(ar(6))
    assert state.dkind == "f"
    state = State(ar(6).astype(np.complex128))
    assert state.dkind == "c"


def test_state_norm():
    state = State(ar(6)).normalize()
    str(state)
    assert len(state) == 1
    assert state.norm()[0] == pytest.approx(1)


def test_state_change_gauge():
    g = geom.graphene(1.42)
    state = State(ar(2, 2).astype(np.complex128), g, gauge="atomic", k=(0.1, 0.2, 0.4))
    assert len(state) == 2
    old = state.state.copy()
    state.change_gauge("lattice")
    assert not np.allclose(old, state.state)
    state.change_gauge("atomic")
    assert np.allclose(old, state.state)


def test_state_change_gauge_nc():
    g = geom.graphene(1.42)
    state = State(ar(2, 4).astype(np.complex128), g, gauge="atomic", k=(0.1, 0.2, 0.4))
    assert len(state) == 2
    old = state.state.copy()
    state.change_gauge("lattice")
    assert not np.allclose(old, state.state)
    state.change_gauge("atomic")
    assert np.allclose(old, state.state)


def test_state_sub():
    state = ar(10, 10)
    state = State(state)
    assert len(state) == 10
    norm = state.norm()
    norm2 = state.norm2()
    for i in range(len(state)):
        assert len(state.sub(i)) == 1
        assert state.sub(i).norm()[0] == norm[i]
        assert state[i].norm()[0] == norm[i]
    for i, sub in enumerate(state):
        assert len(sub) == 1
        assert sub.norm()[0] == norm[i]

    for i, sub in enumerate(state.iter(True)):
        assert (sub**2).sum() == norm2[i]

    assert np.allclose(
        state.sub(np.array([False, True, False, True])).state, state.sub([1, 3]).state
    )

    sub = state.sub(np.array([False, True, False, True]))
    state.sub([1, 3], inplace=True)
    assert np.allclose(sub.state, state.state)


def test_state_outer():
    state = ar(10, 10)
    state = State(state)
    out = state.outer()
    o = out.copy()
    o.fill(0)
    for i, sub in enumerate(state):
        o += outer(sub.state[0, :])

    assert np.allclose(out, o)
    o = state.outer(state)
    assert np.allclose(out, o)


def test_state_outer_matrix():
    state = ar(10, 10)
    M = ar(10)
    state = State(state)
    out = state.outer(matrix=M)
    out_full = state.outer(matrix=np.diag(M))
    assert np.allclose(out, out_full)


def test_state_inner():
    state = ar(10, 10)
    state = State(state)
    inner = state.inner(diag=False)
    assert np.allclose(inner, state.inner(state, projection="matrix"))
    inner_diag = state.inner()
    assert np.allclose(np.diag(inner), inner_diag)


def test_state_inner_matrix():
    state = ar(10, 10)
    M = ar(10)
    state = State(state)
    inner = state.inner(matrix=M, diag=False)
    assert np.allclose(
        inner, state.inner(state, matrix=np.diag(M), projection="matrix")
    )
    inner_diag = state.inner(matrix=M)
    assert np.allclose(np.diag(inner), inner_diag)
    inner_diag = state.inner(matrix=np.diag(M))
    assert np.allclose(np.diag(inner), inner_diag)


def test_state_inner_differing_size():
    state1 = State(ar(8, 10))
    state2 = State(ar(4, 10))

    inner = state1.inner(state2, projection="matrix")
    assert inner.shape == (8, 4)


def test_state_inner_projections():
    g = geom.graphene(atoms=Atom(6, R=(1, 2))) * (2, 2, 1)
    n = g.no + 1
    state = State(ar(n, g.no), parent=g)

    for projs, shape in (
        (("diag", "diagonal", True), (n,)),
        (("trace", "sum"), tuple()),
        (("matrix", False), (n, n)),
        (("basis", "orbitals", "orbital"), (n, g.no)),
        (("atoms", "atom"), (n, g.na)),
    ):
        for proj in projs:
            data = state.inner(projection=proj)
            assert data.shape == shape


def test_state_norm_projections():
    g = geom.graphene(atoms=Atom(6, R=(1, 2))) * (2, 2, 1)
    n = g.no + 1
    state = State(ar(n, g.no), parent=g)
    assert state.shape[0] != state.shape[1]

    for projs, shape in (
        (("diagonal", True), (n,)),
        (("trace", "sum"), tuple()),
        (("hadamard", "basis", "orbitals", "orbital"), (n, g.no)),
        (("atoms", "atom", "hadamard:atoms"), (n, g.na)),
    ):
        for proj in projs:
            data = state.norm2(projection=proj)
            assert data.shape == shape


def test_state_phase_max():
    state = np.random.rand(10, 10) + 1j * np.random.rand(10, 10)
    state1 = State(state)
    state2 = State(-state)
    ph1 = state1.phase()
    ph2 = state2.phase()
    assert np.allclose(ph1, ph2 + np.pi)


def test_state_phase_all():
    state = np.random.rand(10, 10) + 1j * np.random.rand(10, 10)
    state1 = State(state)
    state2 = State(-state)
    ph1 = state1.phase("all")
    ph2 = state2.phase("all")
    assert np.allclose(ph1, ph2 + np.pi)


def test_state_align_phase():
    state = ortho_matrix(10)
    state1 = State(state)
    state2 = State(-state)

    # This should rotate all back
    align2 = state2.align_phase(state1)
    assert np.allclose(state1.state, align2.state)
    align2, idx = state2.align_phase(state1, ret_index=True)
    assert not np.allclose(state1.state, state2.state)
    state2.align_phase(state1, inplace=True)
    assert np.allclose(state1.state, state2.state)
    state2 = State(-state)
    idx = state2.align_phase(state1, inplace=True, ret_index=True)
    assert np.allclose(state1.state, state2.state)


def test_state_ipr():
    state = State(ortho_matrix(10))
    ipr = state.ipr()
    assert ipr.shape == (10,)


def test_state_align_norm():
    state = ortho_matrix(10)
    state1 = State(state)
    idx = np.arange(len(state))
    np.random.shuffle(idx)
    state2 = state1.sub(idx)

    # This should swap all back
    align2 = state2.align_norm(state1)
    assert np.allclose(state1.state, align2.state)
    # This should swap all back
    align2, idx2 = state2.align_norm(state1, ret_index=True)
    assert np.allclose(state1.state, align2.state)
    align1 = state2.sub(idx2)
    assert np.allclose(state1.state, align1.state)


def test_state_align_norm2():
    state = ortho_matrix(10)
    state1 = State(state)
    idx = np.arange(len(state))
    np.random.shuffle(idx)
    state2 = state1.sub(idx)

    # This should swap all back
    align2, idx2 = state2.align_norm(state1, ret_index=True)
    assert np.allclose(state1.state, align2.state)
    assert np.allclose(state1.state, state2.sub(idx2).state)


def test_state_rotate():
    state = State([[1 + 1.0j, 1.0], [0.1 - 0.1j, 0.1]])

    # Angles are 45 and -45
    s = state.copy()
    assert np.pi / 4 == pytest.approx(np.angle(s.state[0, 0]))
    assert 0 == pytest.approx(np.angle(s.state[0, 1]))
    assert -np.pi / 4 == pytest.approx(np.angle(s.state[1, 0]))
    assert 0 == pytest.approx(np.angle(s.state[1, 1]))

    s.rotate(inplace=True)  # individual false
    assert 0 == pytest.approx(np.angle(s.state[0, 0]))
    assert -np.pi / 4 == pytest.approx(np.angle(s.state[0, 1]))
    assert -np.pi / 2 == pytest.approx(np.angle(s.state[1, 0]))
    assert -np.pi / 4 == pytest.approx(np.angle(s.state[1, 1]))

    s = state.copy()
    s.rotate(individual=True, inplace=True)
    assert 0 == pytest.approx(np.angle(s.state[0, 0]))
    assert -np.pi / 4 == pytest.approx(np.angle(s.state[0, 1]))
    assert 0 == pytest.approx(np.angle(s.state[1, 0]))
    assert np.pi / 4 == pytest.approx(np.angle(s.state[1, 1]))

    s = state.copy()
    s = s.rotate(np.pi / 4, individual=True)
    assert np.pi / 4 == pytest.approx(np.angle(s.state[0, 0]))
    assert 0 == pytest.approx(np.angle(s.state[0, 1]))
    assert np.pi / 4 == pytest.approx(np.angle(s.state[1, 0]))
    assert np.pi / 2 == pytest.approx(np.angle(s.state[1, 1]))


def test_cstate_creation():
    state = StateC(ar(6), 1)
    assert len(state) == 1
    state = StateC(ar(6, 6), ar(6))
    assert len(state) == 6
    assert np.allclose(state.c, ar(6))
    str(state)
    state2 = state.copy()
    assert np.allclose(state2.c, state.c)
    assert np.allclose(state2.state, state.state)
    state2 = state.asState()
    assert np.allclose(state2.state, state.state)
    state2 = state.asCoefficient()
    assert np.allclose(state2.c, state.c)


def test_cstate_repr():
    state = StateC(ar(6), 1)
    assert len(state) == 1
    str(state)
    state = StateC(ar(6), 1, parent=geom.graphene())
    str(state)
    assert len(state) == 1


def test_cstate_sub():
    state = StateC(ar(10, 10), ar(10))
    assert len(state) == 10
    norm = state.norm()
    norm2 = state.norm2()
    for i in range(len(state)):
        assert len(state.sub(i)) == 1
        assert state.sub(i).norm()[0] == norm[i]
        assert state[i].norm()[0] == norm[i]
        assert state[i].norm2()[0] == norm2[i]
        assert state[i].c[0] == state.c[i]
    for i, sub in enumerate(state):
        assert len(sub) == 1
        assert sub.norm()[0] == norm[i]

    sub = state.sub(np.array([False, True, False, True]))
    state.sub([1, 3], inplace=True)
    assert np.allclose(sub.c, state.c)
    assert np.allclose(sub.state, state.state)


def test_cstate_sort():
    state = StateC(ar(10, 10), ar(10))
    sort = state.sort()
    assert len(state) == len(sort)
    c = sort.c[::-1]
    sort_descending = sort.sort(False)
    assert np.allclose(c, sort_descending.c)


def test_cstate_norm():
    state = StateC(ar(10, 10), ar(10)).normalize()
    assert len(state) == 10
    assert np.allclose(state.norm(), 1)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest
from pytest import approx

from sisl.io.orca.stdout import *

pytestmark = [pytest.mark.io, pytest.mark.orca]


def test_tags(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)
    assert out.info.na == 2
    assert out.info.no == 62
    assert out.completed()


def test_read_electrons(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    N = out.read_electrons[:]()
    assert N[0, 0] == approx(7.999998537730)
    assert N[0, 1] == approx(6.999998987205)

    N = out.read_electrons[-1]()
    assert N[0] == approx(7.999998537734)
    assert N[1] == approx(6.999998987209)


def test_charge_name(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)
    for name in ["mulliken", "MULLIKEN", "loewdin", "Lowdin", "LÖWDIN"]:
        assert out.read_charge(name=name) is not None


def test_charge_mulliken_atom(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="mulliken", projection="atom")
    S = out.read_charge[:](name="mulliken", projection="atom", spin=True)
    assert len(C) == 2
    assert C[0][0] == approx(0.029160)
    assert S[0][0] == approx(0.687779, abs=1e-4)
    assert C[0][1] == approx(-0.029160, abs=1e-4)
    assert S[0][1] == approx(0.312221, abs=1e-4)
    assert C[1][0] == approx(0.029158, abs=1e-4)
    assert S[1][0] == approx(0.687793, abs=1e-4)
    assert C[1][1] == approx(-0.029158, abs=1e-4)
    assert S[1][1] == approx(0.312207, abs=1e-4)

    C = out.read_charge[-1](name="mulliken", projection="atom")
    S = out.read_charge[-1](name="mulliken", projection="atom", spin=True)
    assert C[0] == approx(0.029158, abs=1e-4)
    assert S[0] == approx(0.687793, abs=1e-4)
    assert C[1] == approx(-0.029158, abs=1e-4)
    assert S[1] == approx(0.312207, abs=1e-4)


def test_lowedin_atom(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="loewdin", projection="atom")
    S = out.read_charge[:](name="loewdin", projection="atom", spin=True)
    assert len(C) == 2
    assert C[0][0] == approx(-0.111221, abs=1e-5)
    assert S[0][0] == approx(0.660316, abs=1e-4)
    assert C[0][1] == approx(0.111221, abs=1e-5)
    assert S[0][1] == approx(0.339684, abs=1e-4)
    assert C[1][0] == approx(-0.111223, abs=1e-5)
    assert S[1][0] == approx(0.660327, abs=1e-4)
    assert C[1][1] == approx(0.111223, abs=1e-4)
    assert S[1][1] == approx(0.339673, abs=1e-4)

    C = out.read_charge[-1](name="loewdin", projection="atom")
    S = out.read_charge[-1](name="loewdin", projection="atom", spin=True)
    assert C[0] == approx(-0.111223, abs=1e-4)
    assert S[0] == approx(0.660327, abs=1e-4)
    assert C[1] == approx(0.111223, abs=1e-4)
    assert S[1] == approx(0.339673, abs=1e-4)


def test_charge_mulliken_reduced(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="mulliken", projection="orbital")
    S = out.read_charge[:](name="mulliken", projection="orbital", spin=True)
    assert len(C) == 2
    # first charge block
    assert C[0][(0, "s")] == approx(3.915850)
    assert C[0][(0, "pz")] == approx(0.710261, abs=1e-5)
    assert C[0][(1, "dz2")] == approx(0.004147, abs=1e-4)
    assert C[0][(1, "p")] == approx(4.116068)
    # first spin block
    assert S[0][(0, "dx2y2")] == approx(0.001163)
    assert S[0][(1, "f+2")] == approx(-0.000122)
    # last charge block
    assert C[1][(0, "pz")] == approx(0.710263)
    assert C[1][(0, "f0")] == approx(0.000681)
    assert C[1][(1, "s")] == approx(3.860487)
    # last spin block
    assert S[1][(0, "p")] == approx(0.685743, abs=1e-4)
    assert S[1][(1, "dz2")] == approx(-0.000163)

    C = out.read_charge[-1](name="mulliken", projection="orbital")
    S = out.read_charge[-1](name="mulliken", projection="orbital", spin=True)
    # last charge block
    assert C[(0, "pz")] == approx(0.710263)
    assert C[(0, "f0")] == approx(0.000681)
    assert C[(1, "s")] == approx(3.860487)
    # last spin block
    assert S[(0, "p")] == approx(0.685743, abs=1e-4)
    assert S[(1, "dz2")] == approx(-0.000163)

    C = out.read_charge[:](name="mulliken", projection="orbital", orbitals="pz")
    assert C[0][0] == approx(0.710261, abs=1e-5)

    S = out.read_charge[:](
        name="mulliken", projection="orbital", orbitals="f+2", spin=True
    )
    assert S[0][1] == approx(-0.000122)

    S = out.read_charge[-1](
        name="mulliken", projection="orbital", orbitals="p", spin=True
    )
    assert S[0] == approx(0.685743, abs=1e-4)


def test_charge_loewdin_reduced(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="loewdin", projection="orbital")
    S = out.read_charge[:](name="loewdin", projection="orbital", spin=True)
    assert len(S) == 2
    assert C[0][(0, "s")] == approx(3.553405)
    assert C[0][(0, "pz")] == approx(0.723111, abs=1e-5)
    assert C[1][(0, "pz")] == approx(0.723113, abs=1e-5)
    assert S[1][(1, "pz")] == approx(-0.010829, abs=1e-4)

    C = out.read_charge[-1](name="loewdin", projection="orbital")
    S = out.read_charge[-1](name="loewdin", projection="orbital", spin=True)
    assert C[(0, "f-3")] == approx(0.017486)
    assert S[(1, "pz")] == approx(-0.010829, abs=1e-4)
    assert C[(0, "pz")] == approx(0.723113, abs=1e-5)
    assert S[(1, "pz")] == approx(-0.010829, abs=1e-4)

    C = out.read_charge[:](name="loewdin", projection="orbital", orbitals="s")
    assert C[0][0] == approx(3.553405)

    C = out.read_charge[-1](name="loewdin", projection="orbital", orbitals="f-3")
    assert C[0] == approx(0.017486)

    C = out.read_charge[-1](name="loewdin", projection="orbital", orbitals="pz")
    assert C[0] == approx(0.723113)


def test_charge_mulliken_full(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="mulliken", projection="orbital", reduced=False)
    S = out.read_charge[:](
        name="mulliken", projection="orbital", reduced=False, spin=True
    )
    assert len(C) == 2
    assert C[0][0] == approx(0.821857)
    assert S[0][0] == approx(-0.000020)
    assert C[0][32] == approx(1.174653)
    assert S[0][32] == approx(-0.000200)
    assert C[1][8] == approx(0.313072, abs=1e-5)
    assert S[1][8] == approx(0.006429, abs=1e-4)

    C = out.read_charge[-1](name="mulliken", projection="orbital", reduced=False)
    S = out.read_charge[-1](
        name="mulliken", projection="orbital", reduced=False, spin=True
    )
    assert C[8] == approx(0.313072, abs=1e-5)
    assert S[8] == approx(0.006429, abs=1e-4)


def test_charge_loewdin_full(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="loewdin", projection="orbital", reduced=False)
    S = out.read_charge[:](
        name="loewdin", projection="orbital", reduced=False, spin=True
    )
    assert len(S) == 2
    assert C[0][0] == approx(0.894846)
    assert S[0][0] == approx(0.000337)
    assert C[0][61] == approx(0.006054)
    assert S[0][61] == approx(0.004362)
    assert C[1][8] == approx(0.312172)
    assert S[1][8] == approx(0.005159, abs=1e-4)

    C = out.read_charge[-1](name="loewdin", projection="orbital", reduced=False)
    S = out.read_charge[-1](
        name="loewdin", projection="orbital", reduced=False, spin=True
    )
    assert C[8] == approx(0.312172)
    assert S[8] == approx(0.005159, abs=1e-4)


def test_charge_atom_unpol(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="mulliken", projection="atom")
    S = out.read_charge[:](name="mulliken", projection="atom", spin=True)
    assert len(C) == 2
    assert S is None
    assert C[0][0] == approx(-0.037652)

    C = out.read_charge[-1](name="mulliken", projection="atom")
    S = out.read_charge[-1](name="mulliken", projection="atom", spin=True)
    assert C[0] == approx(-0.037652)
    assert S is None

    C = out.read_charge[-1](name="loewdin", projection="atom")
    S = out.read_charge[-1](name="loewdin", projection="atom", spin=True)
    assert C[0] == approx(-0.259865)
    assert S is None


def test_charge_orbital_reduced_unpol(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[:](name="mulliken", projection="orbital")
    S = out.read_charge[:](name="mulliken", projection="orbital", spin=True)
    assert len(C) == 2
    assert S is None
    assert C[0][(0, "py")] == pytest.approx(0.534313)
    assert C[1][(1, "px")] == pytest.approx(1.346363)

    C = out.read_charge[-1](name="mulliken", projection="orbital")
    S = out.read_charge[-1](name="mulliken", projection="orbital", spin=True)
    assert C[(0, "px")] == pytest.approx(0.954436, abs=1e-5)
    assert S is None

    C = out.read_charge[-1](name="mulliken", projection="orbital", orbitals="px")
    S = out.read_charge[-1](
        name="mulliken", projection="orbital", orbitals="px", spin=True
    )
    assert C[0] == pytest.approx(0.954436, abs=1e-5)
    assert S is None

    C = out.read_charge[-1](name="loewdin", projection="orbital")
    S = out.read_charge[-1](name="loewdin", projection="orbital", spin=True)
    assert C[(0, "d")] == pytest.approx(0.315910, abs=1e-5)
    assert S is None

    C = out.read_charge[-1](name="loewdin", projection="orbital", orbitals="d")
    S = out.read_charge[-1](
        name="loewdin", projection="orbital", orbitals="d", spin=True
    )
    assert C[0] == pytest.approx(0.315910, abs=1e-5)
    assert S is None


def test_charge_orbital_full_unpol(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule.output")
    out = stdoutSileORCA(f)

    C = out.read_charge[-1](name="mulliken", projection="orbital", reduced=False)
    S = out.read_charge[-1](
        name="mulliken", projection="orbital", reduced=False, spin=True
    )
    assert C is None
    assert S is None


def test_read_energy(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    E = out.read_energy[:]()
    assert len(E) == 2
    assert E[0].total != 0

    E = out.read_energy[-1]()
    assert E.total == approx(-3532.4784695729268)


def test_read_energy_vdw(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule.output")
    out = stdoutSileORCA(f)

    E = out.read_energy[-1]()
    assert E.vdw != 0
    assert E.total == approx(-3081.2640328972802)


def test_read_orbital_energies(sisl_files):
    f = sisl_files("orca", "nitric_oxide", "molecule.output")
    out = stdoutSileORCA(f)

    E = out.read_orbital_energies[:]()
    assert E[0][0, 0] == approx(-513.8983)
    assert E[0][0, 1] == approx(-513.6538)
    assert E[0][61, 0] == approx(1173.4258)
    assert E[0][61, 1] == approx(1173.6985)
    assert E[1][0, 0] == approx(-513.8983)
    assert E[1][0, 1] == approx(-513.6538)
    assert E[1][61, 0] == approx(1173.4259)
    assert E[1][61, 1] == approx(1173.6985)

    E = out.read_orbital_energies[-1]()
    assert E.shape == (out.info.no, 2)
    assert E[61, 0] == approx(1173.4259)


def test_read_orbital_energies_unpol(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule.output")
    out = stdoutSileORCA(f)

    E = out.read_orbital_energies[:]()
    assert E[0][0] == approx(-513.0978)
    assert E[0][61] == approx(1171.5965)
    assert E[1][0] == approx(-513.0976)
    assert E[1][61] == approx(1171.5967)

    E = out.read_orbital_energies[-1]()
    assert E.shape == (out.info.no,)
    assert E[0] == approx(-513.0976)
    assert E[61] == approx(1171.5967)


def test_multiple_calls(sisl_files):
    f = sisl_files("orca", "carbon_monoxide", "molecule.output")
    out = stdoutSileORCA(f)

    N = out.read_electrons[:]()
    assert len(N) == 2

    E = out.read_orbital_energies[:]()
    assert len(E) == 2

    E = out.read_energy[:]()
    assert len(E) == 2

    C = out.read_charge[:](name="mulliken", projection="atom")
    assert len(C) == 2

    N = out.read_electrons[:]()
    assert len(N) == 2



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl.io.vasp.eigenval import *

pytestmark = [pytest.mark.io, pytest.mark.vasp]


def test_read_eigenval(sisl_files):
    f = sisl_files("vasp", "graphene", "EIGENVAL")
    eigs = eigenvalSileVASP(f).read_data()
    eigsHa = eigenvalSileVASP(f).read_data(units="Ha")
    assert not np.allclose(eigs, eigsHa)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

import sisl
from sisl.io.siesta.stdout import *

pytestmark = [pytest.mark.io, pytest.mark.siesta]

# tests here tests charge reads for output
#  voronoi + hirshfeld: test_vh_*
#  voronoi + hirshfeld + mulliken: test_vhm_*
#  voronoi: test_v_*
#  hirshfeld: test_h_*
#  mulliken: test_m_*

Opt = sisl.Opt
SileError = sisl.SileError


def with_pandas():
    try:
        import pandas

        return True
    except ImportError:
        return False


@pytest.mark.parametrize("name", ("voronoi", "Hirshfeld", "mulliken"))
def test_vhm_empty_file(name, sisl_files):
    f = sisl_files("siesta", "ancient", "voronoi_hirshfeld_4.1_none.out")
    out = stdoutSileSiesta(f)

    with pytest.raises(SileError, match="any charges"):
        out.read_charge(name)

    with pytest.raises(SileError, match="any charges"):
        out.read_charge(name, iscf=-1, imd=Opt.NONE)

    with pytest.raises(SileError, match="any charges"):
        out.read_charge(name, iscf=-1, imd=-1)

    with pytest.raises(SileError, match="any charges"):
        out.read_charge(name, iscf=None, imd=-1)


@pytest.mark.parametrize("fname", ("end", "pol_end"))
@pytest.mark.parametrize("name", ("voronoi", "Hirshfeld"))
def test_vh_final(fname, name, sisl_files):
    f = sisl_files("siesta", "charges", fname, "RUN.out")
    out = stdoutSileSiesta(f)

    q = out.read_charge(name, iscf=None, imd=None)
    assert q.size > 0
    assert not all(v is None for v in q)

    q0 = out.read_charge(name, iscf=Opt.NONE, imd=Opt.NONE)
    assert np.allclose(q, q0)

    with pytest.raises(SileError, match="MD/SCF charges"):
        out.read_charge(name, iscf=-1, imd=Opt.NONE)

    with pytest.raises(SileError, match="MD/SCF charges"):
        out.read_charge(name, iscf=-1, imd=-1)

    with pytest.raises(SileError, match="MD/SCF charges"):
        out.read_charge(name, iscf=None, imd=-1)

    if with_pandas():
        df = out.read_charge(name, as_dataframe=True)
        assert np.allclose(df.values, q)


@pytest.mark.parametrize("fname", ("md", "pol_md", ("ancient", "pol_md"), "nc_md"))
@pytest.mark.parametrize("name", ("voronoi", "Hirshfeld", "mulliken"))
def test_vhm_md(name, fname, sisl_files):
    if isinstance(fname, tuple):
        if name == "mulliken":
            # we don't have this in the test, so simply return
            # (this isn't a test after all)
            return
        f = sisl_files("siesta", "ancient", f"voronoi_hirshfeld_4.1_{fname[1]}.out")
    else:
        f = sisl_files("siesta", "charges", fname, "RUN.out")
    out = stdoutSileSiesta(f)

    q = out.read_charge(name)

    q0 = out.read_charge(name, imd=Opt.ALL)
    assert np.allclose(q, q0)

    q0 = out.read_charge(name, imd=-1)
    assert np.allclose(q[-1], q0)

    with pytest.raises(SileError, match="final charges"):
        out.read_charge(name, iscf=None, imd=None)

    for iscf in [-1, Opt.ALL]:
        # space to not match MD/SCF charges
        with pytest.raises(SileError, match=" SCF charges"):
            out.read_charge(name, iscf=-1)

    if with_pandas():
        df = out.read_charge(name, imd=Opt.ALL, as_dataframe=True)
        assert np.allclose(q.ravel(), df.values.ravel())

        df = out.read_charge(name, imd=-1, as_dataframe=True)
        assert np.allclose(q[-1].ravel(), df.values.ravel())


@pytest.mark.parametrize("fname", ("md_scf", "pol_md_scf", "nc_md_scf", "soc_md_scf"))
@pytest.mark.parametrize("name", ("voronoi", "Hirshfeld"))
def test_vh_md_scf(name, fname, sisl_files):
    f = sisl_files("siesta", "charges", fname, "RUN.out")
    out = stdoutSileSiesta(f)

    q = out.read_charge(name)

    with pytest.raises(SileError, match="final charges"):
        out.read_charge(name, iscf=None, imd=Opt.NONE)

    q0 = out.read_charge(name, iscf=-1, imd=-1)
    assert np.allclose(q0, q[-1][-1])

    q0 = out.read_charge(name, iscf=Opt.ANY, imd=-1)
    assert np.allclose(q0, q[-1])

    q0 = out.read_charge(name, iscf=-1, imd=Opt.ALL)
    assert np.allclose(q0, np.stack(tuple(qq[-1] for qq in q)))

    # iscf is
    q0 = out.read_charge(name, iscf=1, imd=Opt.ALL)
    assert np.allclose(q0, np.stack(tuple(qq[1] for qq in q)))

    q0 = out.read_charge(name, iscf=None, imd=-1)
    assert np.allclose(q0, q[-1][-1])

    if with_pandas():
        df = out.read_charge(name, iscf=-1, imd=-1, as_dataframe=True)
        assert np.allclose(df.values, q[-1][-1])



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
import io

import numpy as np
import pytest

import sisl


def test_buffer_write_read():
    """Test that we can write and read using a text buffer."""

    buffer = io.StringIO("")

    geometry = sisl.geom.graphene()

    sisl.io.fdfSileSiesta(buffer).write_geometry(geometry)

    buffer = io.StringIO(buffer.getvalue())

    read_geometry = sisl.io.fdfSileSiesta(buffer).read_geometry()

    assert np.allclose(read_geometry.xyz, geometry.xyz)
    assert np.allclose(read_geometry.cell, geometry.cell)


def test_buffer_write_read_binary():
    """Test that we can write and read binary files using buffers"""

    buffer = io.BytesIO()

    geometry = sisl.geom.graphene()
    H = sisl.Hamiltonian(geometry)
    H[0, 0] = 1.3
    H[1, 1] = 1.5

    sisl.io.hsxSileSiesta(buffer, mode="wb").write_hamiltonian(H)

    buffer = io.BytesIO(buffer.getvalue())

    read_H = sisl.io.hsxSileSiesta(buffer).read_hamiltonian()

    assert np.allclose(H.tocsr().toarray(), read_H.tocsr().toarray())


def test_buffer_write_read_cdf():
    """Test that we can write and read CDF files using buffers"""
    pytest.importorskip("netCDF4")

    buffer = io.BytesIO()

    geometry = sisl.geom.graphene()
    grid = sisl.Grid((2, 2, 2), geometry=geometry)
    grid[:] = np.random.random(grid.shape)

    nc = sisl.io.gridncSileSiesta(buffer, mode="wb")
    nc.write_grid(grid)

    buffer = io.BytesIO(buffer.getvalue())

    read_grid = sisl.io.gridncSileSiesta(buffer).read_grid()

    assert np.allclose(grid.grid, read_grid.grid)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

from functools import partial

import numpy as np
import pytest

from sisl import Geometry, Lattice
from sisl.geom import NeighborFinder
from sisl.geom._neighbors import (
    AtomNeighborList,
    CoordNeighborList,
    CoordsNeighborList,
    FullNeighborList,
    PartialNeighborList,
    UniqueNeighborList,
)

pytestmark = [pytest.mark.geometry, pytest.mark.geom, pytest.mark.neighbor]


tr_fixture = partial(pytest.fixture, scope="module", params=[True, False])


def request_param(request):
    return request.param


sphere_overlap = tr_fixture()(request_param)
multiR = tr_fixture()(request_param)
self_interaction = tr_fixture()(request_param)
post_setup = tr_fixture()(request_param)
pbc = tr_fixture()(request_param)


def set_pbc(geom, _pbc):
    if _pbc:
        geom.lattice.set_boundary_condition(Lattice.BC.PERIODIC)
    else:
        geom.lattice.set_boundary_condition(Lattice.BC.UNKNOWN)


@pytest.fixture(scope="module")
def neighfinder(sphere_overlap, multiR, pbc):
    geom = Geometry([[0, 0, 0], [1.2, 0, 0], [9, 0, 0]], lattice=[10, 10, 7])
    set_pbc(geom, pbc)

    R = np.array([1.1, 1.5, 1.2]) if multiR else 1.5

    neighfinder = NeighborFinder(geom, R=R, overlap=sphere_overlap)

    neighfinder.assert_consistency()

    return neighfinder


@pytest.fixture(scope="module")
def expected_neighs(sphere_overlap, multiR, self_interaction, pbc):
    first_at_neighs = []
    if not (multiR and not sphere_overlap):
        first_at_neighs.append([0, 1, 0, 0, 0])
    if self_interaction:
        first_at_neighs.append([0, 0, 0, 0, 0])
    if pbc:
        first_at_neighs.append([0, 2, -1, 0, 0])

    second_at_neighs = [[1, 0, 0, 0, 0]]
    if self_interaction:
        second_at_neighs.insert(0, [1, 1, 0, 0, 0])
    if pbc and sphere_overlap:
        second_at_neighs.append([1, 2, -1, 0, 0])

    third_at_neighs = []
    if self_interaction:
        third_at_neighs.append([2, 2, 0, 0, 0])
    if pbc:
        if sphere_overlap:
            third_at_neighs.append([2, 1, 1, 0, 0])
        third_at_neighs.append([2, 0, 1, 0, 0])

    return (
        np.array(first_at_neighs),
        np.array(second_at_neighs),
        np.array(third_at_neighs),
    )


def test_neighfinder_setup(sphere_overlap, multiR, post_setup):
    geom = Geometry([[0, 0, 0], [1, 0, 0]], lattice=[10, 10, 7])

    R = np.array([0.9, 1.5]) if multiR else 1.5

    if post_setup:
        # We are going to create a data structure with the wrong parameters,
        # and then update it.
        finder = NeighborFinder(geom, R=R - 0.7, overlap=not sphere_overlap)
        finder.setup(R=R, overlap=sphere_overlap)
    else:
        # Initialize finder with the right parameters.
        finder = NeighborFinder(geom, R=R, overlap=sphere_overlap)

    # Check that R is properly set when its a scalar and an array.
    if multiR:
        assert isinstance(finder.R, np.ndarray)
        assert np.all(finder.R == R)
    else:
        assert finder.R.ndim == 0
        assert finder.R == R

    # Assert that we have stored a copy of the geometry
    assert isinstance(finder.geometry, Geometry)
    assert finder.geometry == geom
    assert finder.geometry is not geom

    # Check that the total number of bins is correct. If sphere_overlap is
    # True, bins are much bigger.
    nbins = (1, 1, 1) if sphere_overlap else (3, 3, 2)
    assert finder.nbins == nbins

    total_bins = 1 if sphere_overlap else 18
    assert finder.total_nbins == total_bins

    # Assert that the data structure is generated.
    for k in ("_list", "_counts", "_heads"):
        assert hasattr(finder, k), k
        assert isinstance(finder._list, np.ndarray), k

    finder.assert_consistency()

    # Check that all bins are empty except one, which contains the two atoms.
    assert (finder._counts == 0).sum() == finder.total_nbins - 1
    assert finder._counts.sum() == 2


def test_neighbor_pairs(neighfinder, self_interaction, expected_neighs):
    neighs = neighfinder.find_neighbors(self_interaction=self_interaction)

    assert isinstance(neighs, FullNeighborList)

    n_neighs = [len(at_neighs) for at_neighs in expected_neighs]
    assert np.all(neighs.n_neighbors == n_neighs)

    for i, (at_neighs, expected_at_neighs) in enumerate(zip(neighs, expected_neighs)):
        assert isinstance(at_neighs, AtomNeighborList)
        assert at_neighs.atom == i
        assert len(expected_at_neighs) == at_neighs.n_neighbors
        if at_neighs.n_neighbors > 0:
            assert np.all(at_neighs.i == expected_at_neighs[:, 0])
            assert np.all(at_neighs.j == expected_at_neighs[:, 1])
            assert np.all(at_neighs.isc == expected_at_neighs[:, 2:])


def test_partial_neighbor_pairs(neighfinder, self_interaction, expected_neighs):
    neighs = neighfinder.find_neighbors(self_interaction=self_interaction, atoms=[1, 2])

    assert isinstance(neighs, PartialNeighborList)

    expected_neighs = expected_neighs[1:]

    n_neighs = [len(at_neighs) for at_neighs in expected_neighs]
    assert np.all(neighs.n_neighbors == n_neighs)

    for i, (at_neighs, expected_at_neighs) in enumerate(zip(neighs, expected_neighs)):
        assert isinstance(at_neighs, AtomNeighborList)
        assert at_neighs.atom == i + 1
        assert len(expected_at_neighs) == at_neighs.n_neighbors
        if at_neighs.n_neighbors > 0:
            assert np.all(at_neighs.i == expected_at_neighs[:, 0])
            assert np.all(at_neighs.j == expected_at_neighs[:, 1])
            assert np.all(at_neighs.isc == expected_at_neighs[:, 2:])


def test_unique_pairs(
    neighfinder, self_interaction, expected_neighs, sphere_overlap, multiR, pbc
):
    # It shouldn't work if you are not requesting sphere overlap and there are
    # multiple cutoff radius.
    if neighfinder.R.ndim == 1 and not neighfinder._overlap:
        with pytest.raises(ValueError):
            neighfinder.find_unique_pairs(self_interaction=self_interaction)
        return

    neighs = neighfinder.find_unique_pairs(self_interaction=self_interaction)

    assert isinstance(neighs, UniqueNeighborList)

    # Convert to a full neighbor list and check that everything is correct.
    full_neighs = neighs.to_full()

    expected_neighs = [
        at_neighs[np.lexsort(at_neighs[:, [1, 0]].T)] if len(at_neighs) > 0 else []
        for at_neighs in expected_neighs
    ]

    assert isinstance(full_neighs, FullNeighborList)

    for at_neighs, expected_at_neighs in zip(full_neighs, expected_neighs):
        assert isinstance(at_neighs, AtomNeighborList)
        print(at_neighs._finder_results, expected_at_neighs)
        assert len(expected_at_neighs) == at_neighs.n_neighbors
        if at_neighs.n_neighbors > 0:
            assert np.all(at_neighs.i == expected_at_neighs[:, 0])
            assert np.all(at_neighs.j == expected_at_neighs[:, 1])
            assert np.all(at_neighs.isc == expected_at_neighs[:, 2:])


def test_close(neighfinder, pbc):
    neighs = neighfinder.find_close([0.3, 0, 0])

    assert isinstance(neighs, CoordsNeighborList)

    first_point_neighs = [[0, 1, 0, 0, 0], [0, 0, 0, 0, 0]]

    if pbc and neighfinder.R.ndim == 0:
        first_point_neighs.append([0, 2, -1, 0, 0])

    expected_neighs = [
        np.array(first_point_neighs),
    ]

    for point_neighs, expected_point_neighs in zip(neighs, expected_neighs):
        assert isinstance(point_neighs, CoordNeighborList)
        assert len(expected_point_neighs) == point_neighs.n_neighbors
        if point_neighs.n_neighbors > 0:
            assert np.all(point_neighs.i == expected_point_neighs[:, 0])
            assert np.all(point_neighs.j == expected_point_neighs[:, 1])
            assert np.all(point_neighs.isc == expected_point_neighs[:, 2:])


def test_close_intcoords(neighfinder):
    """Test the case when the input coordinates are integers.

    (cython routine needs floats)
    """
    neighfinder.find_close([0, 0, 0])


def test_no_neighbors(pbc):
    """Test the case where there are no neighbors, to see that it doesn't crash."""

    geom = Geometry([[0, 0, 0]])
    set_pbc(geom, pbc)

    finder = NeighborFinder(geom, R=1.5)

    neighs = finder.find_neighbors()

    assert isinstance(neighs, FullNeighborList)

    for i, at_neighs in enumerate(neighs):
        assert isinstance(at_neighs, AtomNeighborList)
        assert at_neighs.atom == i
        assert at_neighs.n_neighbors == 0

    neighs = finder.find_unique_pairs()

    assert isinstance(neighs, UniqueNeighborList)
    neighs = neighs.to_full()

    assert isinstance(neighs, FullNeighborList)

    for i, at_neighs in enumerate(neighs):
        assert isinstance(at_neighs, AtomNeighborList)
        assert at_neighs.atom == i
        assert at_neighs.n_neighbors == 0


def test_R_too_big(pbc):
    """Test the case when R is so big that it needs a bigger bin
    than the unit cell."""

    geom = Geometry([[0, 0, 0], [1, 0, 0]], lattice=[2, 10, 10])
    set_pbc(geom, pbc)

    neighfinder = NeighborFinder(geom, R=1.5)

    neighs = neighfinder.find_unique_pairs()

    assert isinstance(neighs, UniqueNeighborList)
    print(neighs._finder_results)

    neighs = neighs.to_full()
    print(neighs._finder_results)

    first_at_neighs = [[0, 1, 0, 0, 0]]
    second_at_neighs = [[1, 0, 0, 0, 0]]

    if pbc:
        first_at_neighs.insert(0, [0, 1, -1, 0, 0])
        second_at_neighs.insert(0, [1, 0, 1, 0, 0])

    expected_neighs = [np.array(first_at_neighs), np.array(second_at_neighs)]

    for at_neighs, expected_at_neighs in zip(neighs, expected_neighs):
        assert isinstance(at_neighs, AtomNeighborList)
        assert len(expected_at_neighs) == at_neighs.n_neighbors
        if at_neighs.n_neighbors > 0:
            assert np.all(at_neighs.i == expected_at_neighs[:, 0])
            assert np.all(at_neighs.j == expected_at_neighs[:, 1])
            assert np.all(at_neighs.isc == expected_at_neighs[:, 2:])

    neighfinder = NeighborFinder(geom, R=[0.6, 2.2], overlap=True)

    neighs = neighfinder.find_close([[0.5, 0, 0]])
    assert isinstance(neighs, CoordsNeighborList)

    expected_neighs = [[0, 1, 0, 0, 0], [0, 0, 0, 0, 0]]
    if pbc:
        expected_neighs.insert(0, [0, 1, -1, 0, 0])
    expected_neighs = [np.array(expected_neighs)]

    for point_neighs, expected_point_neighs in zip(neighs, expected_neighs):
        assert isinstance(point_neighs, CoordNeighborList)
        assert len(expected_point_neighs) == point_neighs.n_neighbors
        if point_neighs.n_neighbors > 0:
            assert np.all(point_neighs.i == expected_point_neighs[:, 0])
            assert np.all(point_neighs.j == expected_point_neighs[:, 1])
            assert np.all(point_neighs.isc == expected_point_neighs[:, 2:])


def test_bin_sizes():
    geom = Geometry([[0, 0, 0], [1, 0, 0]], lattice=[2, 10, 10])

    # We should have fewer bins along the first lattice vector
    n1 = NeighborFinder(geom, R=1.5, bin_size=2)
    n2 = NeighborFinder(geom, R=1.5, bin_size=4)

    assert n1.total_nbins > n2.total_nbins
    # When the bin is bigger than the unit cell, this situation
    # occurs
    assert n1.nbins[0] == n2.nbins[0]
    assert n1.nbins[1] > n2.nbins[1]
    assert n1.nbins[2] > n2.nbins[2]

    # We should have the same number of bins the 2nd and 3rd lattice vectors
    n3 = NeighborFinder(geom, R=1.5, bin_size=2)
    n4 = NeighborFinder(geom, R=1.5, bin_size=(2, 4, 4))

    assert n3.nbins[0] == n4.nbins[0]
    assert n3.nbins[1] > n4.nbins[1]
    assert n3.nbins[2] > n4.nbins[2]


def test_outside_box():
    geom = Geometry([[0, 0, 0], [3, 0, 0]], lattice=[2, 10, 10])

    # The neighbor finder should raise an error if an atom is outside the box
    # of the unit cell, because it is not supported for now.
    # IT SHOULD BE SUPPORTED IN THE FUTURE
    with pytest.raises(ValueError):
        n = NeighborFinder(geom, R=1.1)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import pytest

import sisl
from sisl.viz.plots import atomic_matrix_plot


def test_atomic_matrix_plot():
    pytest.importorskip("skimage", reason="scikit-image not importable")

    graphene = sisl.geom.graphene()
    H = sisl.Hamiltonian(graphene)

    atomic_matrix_plot(H)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

""" pytest test configures """


import numpy as np
import pytest

import sisl

pytestmark = [pytest.mark.io, pytest.mark.siesta]


def test_si_pdos_gamma(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_gamma", "Si_pdos.PDOS.xml"))
    geom, E, pdos = si.read_data()
    assert len(geom) == 2
    assert len(E) == 200
    assert pdos.shape == (1, geom.no, 200)


def test_si_pdos_gamma_xarray(sisl_files):
    pytest.importorskip("xarray", reason="xarray not available")
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_gamma", "Si_pdos.PDOS.xml"))
    X = si.read_data(as_dataarray=True)
    assert len(X.geometry) == 2
    assert len(X.E) == 200
    assert len(X.spin) == 1
    assert X.spin[0] == "sum"
    size = np.prod(X.shape[2:])
    assert size >= X.geometry.no


def test_si_pdos_kgrid(sisl_files):
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.PDOS.xml"))
    geom, E, pdos = si.read_data()
    assert len(geom) == 2
    assert len(E) == 200
    assert pdos.shape == (1, geom.no, 200)


def test_si_pdos_kgrid_xarray(sisl_files):
    pytest.importorskip("xarray", reason="xarray not available")
    si = sisl.get_sile(sisl_files("siesta", "Si_pdos_k", "Si_pdos.PDOS.xml"))
    X = si.read_data(as_dataarray=True)
    assert len(X.geometry) == 2
    assert len(X.E) == 200
    assert len(X.spin) == 1
    assert X.spin[0] == "sum"
    size = np.prod(X.shape[2:])
    assert size >= X.geometry.no



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
from __future__ import annotations

import numpy as np
import pytest

from sisl import Atom, Geometry, Hamiltonian, Lattice
from sisl.physics import yield_manifolds

pytestmark = [pytest.mark.physics, pytest.mark.physics_feature]


def test_yield_manifolds_eigenvalues():
    g = Geometry(
        [[i, 0, 0] for i in range(10)],
        Atom(6, R=1.01),
        lattice=Lattice([10, 1, 5.0], nsc=[3, 3, 1]),
    )
    H = Hamiltonian(g, dtype=np.float64)
    H.construct([(0.1, 1.5), (1.0, 0.1)])

    all_manifolds = []
    for manifold in yield_manifolds(H.eigh()):
        all_manifolds.extend(manifold)

    assert np.allclose(all_manifolds, np.arange(len(H)))

    all_manifolds = []
    for manifold in yield_manifolds(H.tile(2, 0).eigh()):
        all_manifolds.extend(manifold)

    assert np.allclose(all_manifolds, np.arange(len(H) * 2))
